#!/usr/bin/env python
import numpy as np
from pyjams.jams.date2dec import date2dec  # jams
from pyjams.jams.dec2date import dec2date  # jams
import os as os
import re as re


def meteo4slt(sltdir, metfile, p_t_rh, outfile,
              pat='[a-zA-Z0-9]*.slt|[a-zA-Z0-9]*.SLT', delimiter=',',
              skiprows=1, format='ascii'):
    """
        To supply EddyFlux (Kolle & Rebmann, 2007) with meteorological
        data, it is necessary to extract for each available *.slt file the
        corresponding air pressure, air temperature and air relative humidity.
        The module looks in sltdir for available *.slt files and uses the metfile
        for syncronisation


        Definition
        ----------
        meteo4slt(sltdir, metfile, p_t_rh, outfile,
              pat='[a-zA-Z0-9]*.slt|[a-zA-Z0-9]*.SLT', delimiter=',',
              skiprows=1, format='ascii'):

        Input
        -----
        sltdir      str, path of the folder containing the *.slt files
        metfile     str, path of the meteo file
        p_t_rh      list, column number of Pressure, T, rH
        outfile     str, path of the output file


        Optional Input
        --------------
        pat         str, regular expression, describing the name pattern of
                    the *.slt files in the indir folder
        delimiter   str, column delimiter of the meteo file (default=',')
        skiprows    int, number of rows to skip at the beginning of the met file
                    e.g. header lines (default=1)
        format      str, time format of the meteo file (default='ascii')


        Output
        ------
        outfile     file, containing Pressure, T and rH values of the meteo
                    file for each *.slt file


        Restrictions
        ------------
        - assumes site name in slt fielname is only one character
          (e.g. W20133652300.slt AND NOT WULF20133652300.slt)
        - currently only supports format='ascii', nice would be 'eng'
        - does not work for data containning multiple years because of ugly
          doy calculation
        - works only for half hourly *.stl and half hourly meteo data


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Arndt Piayda, Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  AP, Jul 2014
        Modified, MC, Aug 2014 - clean up and Python 3
    """

    # half hour in jd
    halfhjd = 1800./86400.

    ###########################################################################
    # reading slt directory
    dirlist = os.listdir(sltdir)
    sltdates = np.array([])

    ###########################################################################
    # remove all files and folders from list which are not *.slt files and get size
    pat = re.compile(pat)
    for item in dirlist:
        if re.search(pat, item):
            sltdates = np.append(sltdates, item[1:-4])

    # replace time steps which dont fit in half hour interval
    mi = np.array([x[-2:] for x in sltdates])
    mi = np.where(mi.astype(int)<30, '00', '30')
    sltdates = np.array([sltdates[i][:-2]+mi[i] for i in range(mi.size)])

    ###########################################################################
    # load meteo data
    metdata = np.loadtxt(metfile, dtype='|S100', delimiter=delimiter,
                         skiprows=skiprows, usecols=[0]+p_t_rh)

    # get the metdate
    if format == 'ascii':
        # shift met dates one half hour back since slt time stamp marks half
        # hour start but meteo date mark half hour end
        jd = date2dec(ascii=metdata[:,0]) - halfhjd
        fulldate = dec2date(jd, fulldate=True)
        adate    = dec2date(jd, ascii=True)
        doy  = np.ceil(date2dec(ascii=adate)-
                       date2dec(yr=fulldate[0][0],
                                mo=1,dy=1,hr=0,mi=0, sc=0)).astype(int)
        doy = np.where((fulldate[3]==0) & (fulldate[4]==0), doy+1, doy)
        metdates = np.array(['%04i%03i%02i%02i'%(fulldate[0][i], doy[i],
                                                fulldate[3][i], fulldate[4][i])
                             for i in range(jd.size)])
    elif format == 'eng':
        # shift met dates one half hour back since slt time stamp marks half
        # hour start but meteo date mark half hour end
        jd = date2dec(eng=metdata[:,0]) - halfhjd
        fulldate = dec2date(jd, fulldate=True)
        adate    = dec2date(jd, ascii=True)
        doibegin = np.array([date2dec(yr=x, mo=1,dy=1,hr=0,mi=0, sc=0) for x in fulldate[0]])
        doy = np.ceil(date2dec(ascii=adate)-doibegin).astype(int)
        doy = np.where((fulldate[3]==0) & (fulldate[4]==0), doy+1, doy)
        metdates = np.array(['%04i%03i%02i%02i'%(fulldate[0][i], doy[i],
                                                fulldate[3][i], fulldate[4][i])
                             for i in range(jd.size)])
    else:
        raise ValueError('meteo4slt: unknown format!')

    ###########################################################################
    # sync both dates
    mask = np.in1d(metdates, sltdates)

    ###########################################################################
    # write output
    np.savetxt(outfile, metdata[mask,1:], '%s', ',')

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
"""
pyjams own color palettes

.. moduleauthor:: Matthias Cuntz

History
    * Written Apr 2022, Matthias Cuntz,
      added pyjams_amwg

"""

__all__ = ['pyjams_cmaps']


# Toned down version of ncl_amwg w/o pink and rose,
# from Cernusak et al. (New Phyt 2022)
# This is the change from the New Phytology editorial office
# to our plots made with ncl_amwg.
# Must be some kind of conversion between RGB, sRGB or Adobe RGB.
pyjams_amwg = [
    (145, 45, 50),
    (213, 75, 40),
    (229, 153, 42),
    (244, 208, 43),
    (206, 172, 130),
    (239, 220, 182),
    (14, 132, 88),
    (141, 185, 66),
    (0, 152, 163),
    (155, 204, 213),
    (93, 136, 185),
    (36, 91, 153),
    (36, 33, 104),
    (124, 104, 157),
    (0, 0, 0),
    (255, 255, 255) ]
pyjams_amwg = [ (c[0]/255, c[1]/255, c[2]/255) for c in pyjams_amwg ][::-1]

# pyjams color maps
pyjams_cmaps = {
    'pyjams_amwg': pyjams_amwg,
}


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
import numpy as np


__all__ = ['clockplot']


def clockplot(sub, si, sti=None, stierr=None,
              iplot       = None,           # plot number for abc2plot
              usetex      = False,
              dxabc       = 2,              # factor to half the plot size for a,b,c,... labels
              dyabc       = 0.5,            # factor to 0.1 time the plot height for a,b,c,... labels
              dxsig       = 1.23,           # factor to half the plot size for signature
              dysig       = -0.05,          # factor to 0.1 time the plot height for signature
              elwidth     = 1.0,            # errorbar line width
              alwidth     = 1.0,            # axis line width
              glwidth     = 0.5,            # grid line width
              acol        = 'black',        # axis colour
              mcol        = '0.4',          # grid colour
              ecol        = 'black',        # error bar colour
              mcols       = ['0.0', '0.4', '0.4', '0.7', '0.7', '1.0'],      # stack colors
              lcols       = ['None', 'None', 'None', 'None', 'None', '0.0'], # stack bourder colours
              hatches     = [None, None, None, None, None, '//'],            # stack hatching
              llxbbox     = 0.0,            # x-anchor legend bounding box
              llybbox     = 1.15,           # y-anchor legend bounding box
              ymax        = 0.8,            # maximum of y-axis
              ntextsize   = 'medium',       # normal textsize (relative to mpl.rcParams['font.size'])
              bmod        = 0.5,            # fraction of ymax from center to start module colours
              alphamod    = 0.7,            # alpha channel for modules
              fwm         = 0.05,           # module width to remove at sides
              ylabel1     = 1.15,           # position of module names
              ylabel2     = 1.35,           # position of class names
              mtextsize   = 'large',        # textsize of module labels (relative to mpl.rcParams['font.size'])
              bpar        = 0.4,            # fraction of ymax from center to start with parameter bars
              fwb         = [0.7,0.4,0.3],  # fractional width of bars
              plwidth     = 0.5,            # stack border line width
              bplabel     = 0.1,            # fractional distance of ymax of param numbers in centre from 0-line
              ptextsize   = 'medium',       # textsize of param numbers in centre (relative to mpl.rcParams['font.size'])
              space4yaxis = 2,              # space for y-axis (integer)
              ytextsize   = 'medium',       # textsize of y-axis (relative to mpl.rcParams['font.size'])
              dobw        = False,          # True: black & white
              docomp      = True,           # True: Print classification on top of modules
              dosig       = False,          # True: add signature to plot
              dolegend    = False,          # True: add legend to each subplot
              doabc       = False,          # True: add subpanel numbering
              modul         = ['Interception',    'Snow',             'Soil moisture', 'Soil moisture',
                             r'Direct\n runoff', r'Evapo-\n transp.', 'Interflow',     'Percolation',
                             'Routing',         'Geology'],      # module names
              modhalign    = None,          # Horizontal alignment of module names
              modvalign    = None,          # Vertical alignment of module names
              comp        = ['P',               'P',                'S',             'ET',
                             'Q',               'ET',               'S',             'S',
                             'Q',               'S'],            # class names
              pmod        = [ 1,                 8,                  9,               8,
                              1,                 3,                  5,               3,
                              5,                 9],             # # of parameters per class
              cmod        = 'mhm',                               # color scheme chosen ('mhm' or 'noah')
              cmap        = None,                                # Color for each module mod (if not given cmod determines colors)
              saname      = ['Sobol', 'weighted Sobol', 'RMSE'], # stack names
              indexname   = [r'$S_i$', r'$S_{Ti}-S_i$'],           # index legend name
              star        = None,                                # star symbols
              dystar      = 0.95,                                # % of ymax for stars
              scol        = '1.0',                               # star colour
              sfcol       = 'None',                              # star face colour
              ssize       = 3.0,                                 # star size
              swidth      = 1.0,                                 # star edge width
              ssym        = '*',                                 # star symbol
              sig         = 'J Mai & M Cuntz'):                  # signature
    r"""
    The clock plot with modules and up to three index stacks.

    The plot currently defaults to mHM but it can be customized for any model output.


    Definition
    ----------
    def clockplot(sub, si, sti=None, stierr=None,
                  usetex=False,
                  dxabc       = 2,
                  dyabc       = 0.5,
                  dxsig       = 1.23,
                  dysig       = -0.05,
                  elwidth     = 1.0,
                  alwidth     = 1.0,
                  glwidth     = 0.5,
                  acol        = 'black',
                  mcol        = '0.4',
                  ecol        = 'black',
                  mcols       = ['0.0', '0.4', '0.4', '0.7', '0.7', '1.0'],
                  lcols       = ['None', 'None', 'None', 'None', 'None', '0.0'],
                  hatches     = [None, None, None, None, None, '//'],
                  llxbbox     =  0.0,
                  llybbox     = 1.15,
                  ymax = 0.8,
                  ntextsize   = 'medium',
                  bmod        = 0.5,
                  alphamod    = 0.7,
                  fwm         = 0.05,
                  ylabel1     = 1.15,
                  ylabel2     = 1.35,
                  mtextsize   = 'large',
                  bpar        = 0.4,
                  fwb         = [0.7,0.4,0.3],
                  plwidth     = 0.5,
                  bplabel     = 0.1,
                  ptextsize   = 'medium',
                  space4yaxis = 2,
                  ytextsize   = 'medium',
                  dobw      = False,
                  docomp    = True,
                  dosig     = False,
                  dolegend  = False,
                  doabc     = False,
                  modul   = ['Interception',    'Snow',             'Soil moisture', 'Soil moisture',
                           'Direct\n runoff', 'Evapo-\n transp.', 'Interflow',     'Percolation',
                           'Routing',         'Geology'],
                  modhalign    = None,
                  modvalign    = None,
                  comp  = ['P',               'P',                'S',             'ET',
                           'Q',               'ET',               'S',             'S',
                           'Q',               'S'],
                  pmod  = [ 1,                 8,                  9,               8,
                            1,                 3,                  5,               3,
                            5,                 9],
                  cmod  = 'mhm',
                  cmap  = [['reds8',   1], ['reds8',   2], ['reds8',   3], ['reds8',   4],
                           ['blues8',  3], ['blues8',  4], ['blues8',  5], ['blues8',  6], 
                           ['reds8',   5], ['reds8',   6]],
                  saname = ['Sobol', 'weighted Sobol', 'RMSE'],
                  indexname = ['$S_i$', '$S_{Ti}-S_i$'],
                  star = None,
                  dystar = 0.95,
                  scol = '1.0',
                  sfcol = 'None',
                  ssize = 3.0,
                  swidth = 1.0,
                  ssym = '*',
                  sig = 'J Mai & M Cuntz'):


    Input
    -----
    sub                          axes handle from for example
                                 sub = fig.add_axes(jams.position(nrow,ncol,iplot), polar=True)
    si                           list, list of arrays, 1D-, or 2D-array si[nstacks, nparameters] of, e.g.,
                                 first-order Sobol' indexes or other indexes


    Optional Input
    --------------
    sti = None                   list, list of arrays, 1D-, or 2D-array sti[nstacks, nparameters] of, e.g.,
                                 total-order Sobol' indexes or upper stack values
    stierr = None                list, list of arrays, 1D-, or 2D-array si[nstacks, nparameters] of
                                 error bars of upper stack indexes
    usetex = False               True: use LaTeX rendering
    dxabc = 2                    % of (max-min) shift to the right from left y-axis for a,b,c,... labels
    dyabc = 0.5                  % of (max-min) shift up from lower x-axis for a,b,c,... labels
    dxsig = 1.23                 % of (max-min) shift to the right from left y-axis for signature
    dysig = -0.05                % of (max-min) shift up from lower x-axis for signature
    elwidth = 1.0                errorbar line width
    alwidth = 1.0                axis line width
    glwidth = 0.5                grid line width
    acol = 'black'               axis colour
    mcol = '0.4'                 grid colour
    ecol = 'black',              error bar colour
    mcols = ['0.0', '0.4', '0.4', '0.7', '0.7', '1.0']        stack colors
    lcols = ['None', 'None', 'None', 'None', 'None', '0.0']   stack border colours
    hatches = [None, None, None, None, None, '//']            stack hatching
    llxbbox =  0.0               x-anchor legend bounding box
    llybbox = 1.15,              y-anchor legend bounding box
    ymax = 0.8                   y-axis maximum
    ntextsize = 'medium'         normal textsize
    bmod = 0.5                   fraction of ymax to start module colours
    alphamod = 0.7               alpha channel for modules
    fwm = 0.05                   module width to remove at sides for space between modules
    ylabel1 = 1.15               fractional position of module names
    ylabel2 = 1.35               fractional position of class names
    mtextsize = 'large'          textsize of module labels
    bpar = 0.4                   fraction of ymax to start parameter bars
    fwb = [0.7,0.4,0.3]          fractional width of bars depending on number of index stacks
    plwidth = 0.5                stack border line width
    bplabel = 0.1                fractional distance of ymax of param numbers in centre from 0-line
    ptextsize = 'medium'         textsize of param numbers in centre
    space4yaxis = 2              space for y-axis top tickmark label (integer)
    ytextsize = 'medium'         textsize of y-axis tickmark labels
    dobw = False                 True: black & white; False: colour
    docomp = True                True: Print classification on top of modules
    dosig = False                True: add signature (sig) to plot
    dolegend = False             True: add legend to each subplot
    doabc = False                True: add subpanel numbering
    modul  = ['Interception',    'Snow',             'Soil moisture', 'Soil moisture',
            'Direct\n runoff', 'Evapo-\n transp.', 'Interflow',     'Percolation',
            'Routing',         'Geology'],                                           module names
    modhalign = None,             list of horizontal alignments of module names
    modvalign = None,             list of vertical alignments of module names
    comp = ['P',               'P',                'S',             'ET',
            'Q',               'ET',               'S',             'S',
            'Q',               'S'],                                                 class names
    pmod = [ 1,                 8,                  9,               8,
             1,                 3,                  5,               3,
             5,                 9],                                                  number of parameters per class
    cmod = 'mhm',                                                                    color scheme chose ('mhm' or 'noah')
    cmap = [['reds8',   1], ['reds8',   2], ['reds8',   3], ['reds8',   4],
             ['blues8',  3], ['blues8',  4], ['blues8',  5], ['blues8',  6], 
             ['reds8',   5], ['reds8',   6]],                                        colors for each class
    saname = ['Sobol', 'weighted Sobol', 'RMSE'],                                    stack names
    indexname = ['$S_i$', '$S_{Ti}-S_i$'],                                           index legend name
    star = None,                                                                     star symbols
    dystar = 0.95,                                                                   % of ymax for stars
    scol = '1.0',                                                                    star colour
    sfcol = 'None',                                                                  star face colour
    ssize = 3.0,                                                                     star size
    swidth = 1.0,                                                                    star edge width
    ssym = '*',                                                                      star symbol
    sig = 'J Mai & M Cuntz'                                                          signature


    Output
    ------
    clockplot on axes sub


    Restrictions
    ------------
    None


    Examples
    --------
    see below if __name__ == '__main__':


    History
    -------
    Written,  MC, JM, AP, Oct 2014

    """
    # Check si[nstacks, nparams]
    if sti is not None:
        assert np.shape(si) == np.shape(sti), 'Lower and upper stacks must have same dimensions.'
    if stierr is not None:
        assert np.shape(si) == np.shape(stierr), 'Error bars must have must have same dimensions as upper indexes.'
    si_shape = np.shape(si)
    if np.size(si_shape) == 1:
        nsi = 1
        isi  = np.array(si)
        isi  = isi[np.newaxis, :]
        if sti is not None:
            isti = np.array(sti)
            isti = isti[np.newaxis, :]
        if stierr is not None:
            istierr = np.array(stierr)
            istierr = istierr[np.newaxis, :]
    elif np.size(si_shape) == 2:
        nsi = si_shape[0]
        if nsi > 3:
            raise ValueError('first data dimension must be <= 3, i.e. at most 3 stacks per parameter supported.')
        isi  = np.array(si)
        if sti is not None:
            isti = np.array(sti)
        if stierr is not None:
            istierr = np.array(stierr)
    else:
        raise ValueError('input data must be 1D or 2D.')
    npar = isi.shape[1]
    if sti is not None:
        idsi = isti - isi

    # alignement of module names
    if modhalign is not None:
        if isinstance(modhalign, (list, tuple, np.ndarray)):
            assert len(modul) == len(modhalign), 'modhalign must be scalar or same size as modul.'
            imodhalign = modhalign
        else:
            imodhalign = [modhalign] * len(modul)
    else:
        imodhalign = ['center'] * len(modul)
    if modvalign is not None:
        if isinstance(modvalign, (list, tuple, np.ndarray)):
            assert len(modul) == len(modvalign), 'modvalign must be scalar or same size as modul.'
            imodvalign = modvalign
        else:
            imodvalign = [modvalign] * len(modul)
    else:
        imodvalign = ['center'] * len(modul)

    # Prepare annotations
    if type(saname) is list:
        isaname = saname[:]
    else:
        isaname = [saname]
    if type(modul) is list:
        ismod = modul[:]
    else:
        ismod = [modul]
    if usetex:
        imod = []
        for ii, i in enumerate(ismod):
            if r'\n' in i:
                ss = i.split(r'\n')
            else:
                ss = [i]
            ss = [ r'' + j.strip() for j in ss ]
            for j, s in enumerate(ss):
                ss[j] = r'$\mathrm{' + s + r'}$'
                if '-' in ss[j]:
                    ss[j] = ss[j].replace('-}$', '}$-')
                if ' ' in ss[j]:
                    ss[j] = ss[j].replace(' ', r'\ ')
                if "'" in isaname[j]:
                    ss[j] = ss[j].replace("'", r"}\\textrm{'}\mathrm{")
            imod.append(ss)
        ismod = imod
        comp = [ r'$\mathbf{' + i + r'}$' for i in comp ]
        for j in range(len(comp)):
            if "'" in comp[j]:
                comp[j] = comp[j].replace("'", r"}\\textrm{'}\mathbf{")
        for j, s in enumerate(isaname):
            isaname[j] = r'$\mathrm{' + s + r'}$'
            if '-' in isaname[j]:
                isaname[j] = isaname[j].replace('-}$', '}$-')
            if ' ' in isaname[j]:
                isaname[j] = isaname[j].replace(' ', r'\ ')
            if "'" in isaname[j]:
                isaname[j] = isaname[j].replace("'", r"}\\textrm{'}\mathrm{")
    else:
        imod = []
        for i in ismod:
            if '\n' in i:
                ss = i.split('\n')
            else:
                ss = [i]
            ss = [ r'' + j.strip() for j in ss ]
            imod.append(ss)
        ismod = imod

    # numbers
    nmod   = len(ismod)  # number modules
    nparam = sum(pmod)   # number of parameters
    param  = np.arange(nparam) + 1
    assert nparam == npar, 'si.shape[1] must be sum(pmod).'
    if star is not None:
        assert nparam == np.size(star), 'There must be the same number of star indications as parameters.'

    # colours
    if dobw:
        c = np.linspace(0.2, 0.85, nmod)
        c = np.ones(nmod) * 0.7
        c = [ str(i) for i in c ]
    else:
        from pyjams.jams.brewer import get_brewer
        if (cmod == 'mhm'):
            c = [get_brewer('rdylbu11', rgb=True)[0],  # interception
                 get_brewer('rdylbu11', rgb=True)[1],  # snow
                 get_brewer('rdylbu11', rgb=True)[2],  # soil moisture
                 get_brewer('rdylbu11', rgb=True)[2],  # soil moisture
                 get_brewer('rdylbu11', rgb=True)[3],  # direct runoff
                 get_brewer('rdylbu11', rgb=True)[4],  # Evapotranspiration
                 get_brewer('rdylbu11', rgb=True)[6],  # interflow
                 get_brewer('rdylbu11', rgb=True)[7],  # percolation
                 get_brewer('rdylbu11', rgb=True)[8],  # routing
                 get_brewer('rdylbu11', rgb=True)[9]]  # geology
        elif (cmod == 'noah'):
            c = [get_brewer('reds8', rgb=True)[3], #  Radiation
                 get_brewer('ylorbr4', rgb=True)[2], #  SoilPhysiology
                 get_brewer('ylorrd8', rgb=True)[2], #  Transfer
                 get_brewer('greens4', rgb=True)[2], #  VegetationStructure
                 get_brewer('bugn4', rgb=True)[2], #  Physiology
                 get_brewer('blues4', rgb=True)[2], #  SoilWater
                 get_brewer('blues6', rgb=True)[5], #  Runoff
                 get_brewer('greys8', rgb=True)[2], #  SnowEnergy
                 get_brewer('greys8', rgb=True)[4], #  SoilEnergy
                 get_brewer('greys8', rgb=True)[5], #  Carbon
                 get_brewer('rdpu5', rgb=True)[2], #  VOC
                 get_brewer('reds8', rgb=True)[5], #  Input*
                 get_brewer('reds8', rgb=True)[3], #  Radiation*
                 get_brewer('ylorbr4', rgb=True)[2], #  SoilPhysiology*
                 get_brewer('ylorrd8', rgb=True)[2], #  Transfer*
                 get_brewer('greens4', rgb=True)[2], #  VegetationStructure*
                 get_brewer('bugn4', rgb=True)[2], #  Physiology*
                 get_brewer('blues4', rgb=True)[2], #  SoilWater*
                 get_brewer('blues6', rgb=True)[5], #  Runoff*
                 get_brewer('blues4', rgb=True)[3], #  SnowWater*
                 get_brewer('greys8', rgb=True)[2], #  SnowEnergy*
                 get_brewer('greys8', rgb=True)[4], #  SoilEnergy*
                 get_brewer('greys8', rgb=True)[5], #  Carbon*
                 get_brewer('greys9', rgb=True)[6]] #  ????
        else:
            if cmap is None:
                raise ValueError("cmod can be only 'mhm' or 'noah', otherwise cmap has to be given.")

    if cmap is not None:
        c = cmap

    # conversion factor from parameter number to radian
    n2rad = 2. * np.pi / (nparam + space4yaxis)

    # -------------------------------------------------------------------------
    # Plot
    #

    # Axes have to be defined externally
    # sub = fig.add_axes(jams.position(nrow,ncol,iplot,hspace=hspace,vspace=vspace), polar=True)
    sub.set_theta_zero_location('N') # 0 is North
    sub.set_theta_direction(-1)      # clockwise

    xlim = [0, 2. * np.pi]
    ylim = [-bpar * ymax, ymax]

    # coloured modules
    mleft   = (space4yaxis + np.cumsum([0] + pmod[:-1]) + fwm) * n2rad  # left start at space4yaxis
    mheight = np.ones(nmod) * ymax * (1.-bmod)                     # height from bmod*ymax to ymax
    mwidth  = (np.array(pmod) - 2. * fwm) * n2rad                    # width is number of params per module
    iidx = np.where(np.array(pmod) > 0)
    bar1    = sub.bar(mleft[iidx], mheight[iidx], mwidth[iidx], bottom=bmod*ymax,
                      color=np.array(c)[iidx], alpha=alphamod,
                      linewidth=0)

    # module and class labels
    xm = mleft + 0.5 * mwidth
    for i in range(nmod):
        if (pmod[i] > 0):
            # module
            nm = len(ismod[i])
            for j, m in enumerate(ismod[i]):
                mlabel12 = (ylabel2 - ylabel1) * 1. / float(nm)
                if (xm[i] < 0.5 * np.pi) | (xm[i] > 1.5 * np.pi):
                    ylab = ylabel2 - (j + 1) * mlabel12
                    rot  = np.rad2deg(-xm[i])
                else:
                    ylab = ylabel2 - (nm - j) * mlabel12
                    rot  = np.rad2deg(-xm[i]) + 180.
                label = sub.text(xm[i], ylab * ymax, m, rotation=rot,
                                 fontsize=ntextsize,
                                 horizontalalignment=imodhalign[i],
                                 verticalalignment=imodvalign[i])
            # class
            if docomp:
                label = sub.text(xm[i], ylabel2 * ymax, comp[i],
                                 fontsize=mtextsize, fontweight='bold',
                                 horizontalalignment='center',
                                 verticalalignment='center')
                if (xm[i] < 0.5 * np.pi) | (xm[i] > 1.5 * np.pi):
                    label.set_rotation(np.rad2deg(-xm[i]))
                else:
                    label.set_rotation(np.rad2deg(-xm[i])+180.)

    # y-axis
    # grid
    nyticks = 5
    dyy     = np.linspace(0, ymax, nyticks)
    gy      = np.delete(dyy[1:-1], nyticks // 2 - 1)
    npoints = 100
    gxx     = np.linspace(0, 2. * np.pi, npoints)
    for i in range(gy.size):
        grid = sub.plot(gxx, np.ones(npoints) * gy[i], linestyle='--',
                        color=mcol, linewidth=glwidth)
    # in "axis normal coordinates" for rectangular ticks, etc.
    dyy    = np.array([0, ymax])
    yy     = ((1. + 2. * bpar) * ymax + dyy) / ((2. + 2. * bpar) * ymax)
    yaxis  = sub.plot([0.5, 0.5], yy, transform=sub.transAxes,
                      linestyle='-', linewidth=alwidth, color=acol)
    nyticks = 5
    xx      = np.ones(nyticks)*0.5
    dyy     = np.linspace(0,ymax,nyticks)
    yy      = ((1.+2.*bpar)*ymax+dyy)/((2.+2.*bpar)*ymax)
    ytickwidth = 0.015
    for i in range(nyticks):
        yticks = sub.plot([xx[i],xx[i]+ytickwidth], [yy[i],yy[i]], transform=sub.transAxes,
                          linestyle='-', linewidth=alwidth, color=acol)
    # y-tickmarks at top and bottom
    tx = xx[0]+1.5*ytickwidth
    ty = yy[0]
    if usetex:
        tt = r'$\mathrm{0}$'
    else:
        tt = '0'
    label = sub.text(tx, ty, tt, transform=sub.transAxes,
                     fontsize=ytextsize, horizontalalignment='left', verticalalignment='bottom')
    bx = xx[-1]+1.5*ytickwidth
    by = yy[-1]
    if usetex:
        bt = r'$\mathrm{'+str(ymax)+'}$'
    else:
        bt = str(ymax)
    label = sub.text(bx, by, bt, transform=sub.transAxes,
                     fontsize=ytextsize, horizontalalignment='left', verticalalignment='top')

    # param numbers in center == x-tickmarks
    xticknames = [ str(i) for i in param[1::4] ]
    if usetex:
        xticknames = [ r'$\mathrm{'+i+'}$' for i in xticknames ]
    shiftx  = np.floor(1./(nsi+1)*10.)/10.
    pleft   = (param+(space4yaxis-1)+shiftx-0.5*fwb[nsi-1]+(nsi-1)/2.*fwb[nsi-1])*n2rad # center at middle of all stacks
    pwidth  = np.ones(nparam)*fwb[nsi-1]*n2rad
    tx = pleft[1::4]+0.5*pwidth[1::4]
    ty = -np.ones(tx.size)*bplabel*ymax
    for i in range(tx.size):
        if (tx[i] < np.pi):
            rot = np.rad2deg(-tx[i])+90.
        else:
            rot = np.rad2deg(-tx[i])-90.
        label = sub.text(tx[i], ty[i], xticknames[i], rotation=rot, fontsize=ptextsize,
                         horizontalalignment='center', verticalalignment='center')

    # Index stacks
    for n in range(nsi):
        # params on bottom
        pleft = (param+(space4yaxis-1)+shiftx-0.5*fwb[nsi-1]+n*fwb[nsi-1])*n2rad # center at 0.5 from param number
        pheight = isi[n,:]
        bar2    = sub.bar(pleft, pheight, pwidth, bottom=0.,
                          facecolor=mcols[2*n], hatch=hatches[2*n], edgecolor=lcols[2*n],
                          linewidth=plwidth)
        if sti is not None:
            # params on top
            pheight = idsi[n,:]
            pbottom = isi[n,:]
            bar2    = sub.bar(pleft, pheight, pwidth, bottom=pbottom,
                              facecolor=mcols[2*n+1], hatch=hatches[2*n+1], edgecolor=lcols[2*n+1],
                              linewidth=plwidth)
        if stierr is not None:
            # error bars
            xx      = pleft+0.5*pwidth
            if sti is not None:
                yy      = isti[n,:]
            else:
                yy      = isi[n,:]
            perr    = istierr[n,:]
            pewidth = 0.1*n2rad
            for i in range(xx.size):
                yerrm = sub.plot([xx[i],xx[i]], [yy[i]-perr[i],yy[i]+perr[i]], # middle line
                                 linestyle='-', linewidth=elwidth, color=ecol)
                yerrl = sub.plot([xx[i]-pewidth,xx[i]+pewidth], [yy[i]-perr[i],yy[i]-perr[i]], # lower bar
                                 linestyle='-', linewidth=elwidth, color=ecol)
                yerru = sub.plot([xx[i]-pewidth,xx[i]+pewidth], [yy[i]+perr[i],yy[i]+perr[i]], # upper bar
                                 linestyle='-', linewidth=elwidth, color=ecol)

    # Stars
    if star is not None:
        pleft = (param+(space4yaxis-1)+shiftx-0.5*fwb[nsi-1]+(nsi-1)/2.*fwb[nsi-1])*n2rad # same as xticklabels
        xstar = (pleft+0.5*pwidth)[star.astype(np.bool)]
        ystar = np.ones(xstar.shape[0]) * ymax * dystar
        star_mucm = sub.plot(xstar, ystar, linestyle='none',
                             marker=ssym, markeredgecolor=scol, markerfacecolor=sfcol,
                             markersize=ssize, markeredgewidth=swidth)

    # subplot numbering
    if doabc and (iplot is not None):
        from pyjams.abc2plot import abc2plot
        abc2plot(sub, dxabc, dyabc, iplot, transform=sub.transAxes,
                 lower=True, parenthesis='close',
                 bold=True, xlarge=True,
                 mathrm=True, usetex=usetex,
                 horizontalalignment='right', verticalalignment='bottom')

    # Signature
    if dosig:
        from pyjams.signature2plot import signature2plot
        signature2plot(sub, dxsig, dysig, sig, transform=sub.transAxes,
                       italic=True, small=True, mathrm=True, usetex=usetex)

    # General settings
    sub.set_xlim(xlim)
    sub.set_ylim(ylim)     # start at -bpar to get offset for bars
    sub.grid(False)
    sub.set_frame_on(False)
    sub.set_xticks([])
    sub.set_xticklabels([])
    sub.set_yticks([])
    sub.set_yticklabels([])

    if dolegend:
        # Fake subplot for legend and numbering
        spos = sub.get_position()
        fig  = sub.get_figure()
        lsub = fig.add_axes([spos.x0+llxbbox*spos.width, spos.y0+llybbox*spos.height, 0.5*spos.width, 0.1*spos.height])

        x1, y1 = lsub.transData.transform_affine(np.array([0,0]))
        x2, y2 = lsub.transData.transform_affine(np.array([1,1]))
        dpi = lsub.figure.dpi                         # pixels per inch
        import matplotlib as mpl
        ss  = mpl.rcParams['font.size']               # text size in points: 1 pt = 1/72 inch
        shifty = 1.0 * ss/72.*float(dpi)/float(y2-y1) # shift by 1.0 of textsize
        for n in range(nsi):
            dy1 = 1. - (n+1.)   * shifty
            dy2 = 1. - (n+0.01) * shifty
            # colour bar
            lsub.fill_between([0,0.3],   [dy1,dy1], [dy2,dy2], linewidth=plwidth, clip_on=False,
                              facecolor=mcols[2*n],   edgecolor=lcols[2*n], hatch=hatches[2*n])
            dx1 = 0.35
            if sti is not None:
                lsub.fill_between([0.3,0.6], [dy1,dy1], [dy2,dy2], linewidth=plwidth, clip_on=False,
                                  facecolor=mcols[2*n+1], edgecolor=lcols[2*n+1], hatch=hatches[2*n+1])
                dx1 = 0.65
            # Annotation to the right
            lsub.text(dx1, 0.5*(dy1+dy2), isaname[n], fontsize=ntextsize,
                      horizontalalignment='left', verticalalignment='center')
        # Annotation on top
        if type(indexname) is list:
            iindexname = indexname[:]
        else:
            iindexname = [indexname]
        t = iindexname[0]
        dy2 = 1. - 0.01 * shifty
        lsub.text(0.15, dy2+0.05, t, fontsize=ntextsize,
                  horizontalalignment='center', verticalalignment='bottom')
        if sti is not None:
            t = iindexname[1]
            lsub.text(0.45, dy2+0.05, t, fontsize=ntextsize,
                      horizontalalignment='center', verticalalignment='bottom')
        # Star
        if star is not None:
            if sti is not None:
                dxstar1 = 0.3
            else:
                dxstar1 = 0.15
            dystar1 = 1. - (2.*nsi+1.01)/2. * shifty
            lsub.plot([dxstar1], [dystar1], linestyle='none', clip_on=False,
                      marker=ssym, markeredgecolor=scol, markerfacecolor=sfcol,
                      markersize=ssize, markeredgewidth=swidth)
            lsub.text(dx1, dystar1, isaname[nsi], fontsize=ntextsize,
                      horizontalalignment='left', verticalalignment='center')

        lsub.set_xlim([0,1])
        lsub.set_ylim([0,1])

        lsub.set_title('')
        lsub.set_xlabel('')
        lsub.set_ylabel('')
        lsub.set_xticks([])
        lsub.set_yticks([])
        lsub.set_axis_off()


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # nn = 52
    # ymax = 0.7
    # si  = np.arange(nn)/np.float(nn-1) * ymax * 0.5
    # sti = np.arange(nn)/np.float(nn-1) * ymax
    # stierr = 0.1*sti

    # pngbase = 'clockplot_'
    # pdffile = '' # 'clockplot.pdf'
    # usetex  = True

    # if (pdffile != '') & (pngbase != ''):
    #     print('\nError: PDF and PNG are mutually exclusive. Only either -p or -g possible.\n')
    #     parser.print_usage()
    #     import sys
    #     sys.exit()

    # if (pdffile == ''):
    #     if (pngbase == ''):
    #         outtype = 'x'
    #     else:
    #         outtype = 'png'
    # else:
    #     outtype = 'pdf'

    # # Main plot
    # nrow        = 3           # # of rows of subplots per figure
    # ncol        = 2           # # of columns of subplots per figure
    # hspace      = 0.05        # x-space between subplots
    # vspace      = 0.09        # y-space between subplots
    # right       = 0.9         # right space on page
    # textsize    = 6           # standard text size
    # dxabc       = 2           # % of (max-min) shift to the right from left y-axis for a,b,c,... labels
    # dyabc       = 0.5           # % of (max-min) shift up from lower x-axis for a,b,c,... labels
    # dxsig       = 1.23        # % of (max-min) shift to the right from left y-axis for a,b,c,... labels
    # dysig       = -0.05       # % of (max-min) shift up from lower x-axis for a,b,c,... labels

    # lwidth      = 1.5         # linewidth
    # elwidth     = 1.0         # errorbar line width
    # alwidth     = 1.0         # axis line width
    # glwidth     = 0.5         # grid line width
    # msize       = 1.0         # marker size
    # mwidth      = 1.0         # marker edge width
    # mcol1       = '0.0'       # primary marker colour
    # mcol2       = '0.4'                     # secondary
    # mcol3       = '0.0' # third
    # mcols       = ['0.0', '0.4', '0.4', '0.7', '0.7', '1.0']
    # lcol1       = colours('blue')   # primary line colour
    # lcol2       = '0.4'
    # lcol3       = '0.0'
    # lcols       = ['None', 'None', 'None', 'None', 'None', '0.0']
    # hatches     = [None, None, None, None, None, '//']

    # # Legend
    # llxbbox     =  0.0        # x-anchor legend bounding box
    # llybbox     = 1.15       # y-anchor legend bounding box
    # llrspace    = 0.          # spacing between rows in legend
    # llcspace    = 1.0         # spacing between columns in legend
    # llhtextpad  = 0.4         # the pad between the legend handle and text
    # llhlength   = 1.5         # the length of the legend handles
    # frameon     = False       # if True, draw a frame around the legend. If None, use rc

    # # PNG
    # dpi           = 600
    # transparent   = False
    # bbox_inches   = 'tight'
    # pad_inches    = 0.035

    # # Clock options
    # ymax = 0.8
    # ntextsize   = 'medium'       # normal textsize
    # # modules
    # bmod        = 0.5            # fraction of ymax from center to start module colours
    # alphamod    = 0.7            # alpha channel for modules
    # fwm         = 0.05           # module width to remove at sides
    # ylabel1     = 1.15           # position of module names
    # ylabel2     = 1.35           # position of class names
    # mtextsize   = 'large'        # 1.3*textsize # textsize of module labels
    # # bars
    # bpar        = 0.4            # fraction of ymax from center to start with parameter bars
    # fwb         = [0.7,0.4,0.3]  # width of bars
    # plwidth     = 0.5
    # # parameters in centre
    # bplabel     = 0.1            # fractional distance of ymax of param numbers in centre from 0-line
    # ptextsize   = 'medium'       # 'small' # 0.8*textsize # textsize of param numbers in centre
    # # yaxis
    # space4yaxis = 2              # space for y-axis (integer)
    # ytextsize   = 'medium'       # 'small' # 0.8*textsize # textsize of y-axis

    # import matplotlib as mpl
    # if (outtype == 'pdf'):
    #     mpl.use('PDF') # set directly after import matplotlib
    #     import matplotlib.pyplot as plt
    #     from matplotlib.backends.backend_pdf import PdfPages
    #     # Customize: http://matplotlib.sourceforge.net/users/customizing.html
    #     mpl.rc('ps', papersize='a4', usedistiller='xpdf') # ps2pdf
    #     mpl.rc('figure', figsize=(8.27,11.69)) # a4 portrait
    #     if usetex:
    #         mpl.rc('text', usetex=True)
    #     else:
    #         #mpl.rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
    #         mpl.rc('font',**{'family':'serif','serif':['times']})
    #     mpl.rc('text.latex', unicode=True)
    # elif (outtype == 'png'):
    #     mpl.use('Agg') # set directly after import matplotlib
    #     import matplotlib.pyplot as plt
    #     mpl.rc('figure', figsize=(8.27,11.69)) # a4 portrait
    #     if usetex:
    #         mpl.rc('text', usetex=True)
    #     else:
    #         #mpl.rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
    #         mpl.rc('font',**{'family':'serif','serif':['times']})
    #     mpl.rc('text.latex', unicode=True)
    #     mpl.rc('savefig', dpi=dpi, format='png')
    # else:
    #     import matplotlib.pyplot as plt
    #     mpl.rc('figure', figsize=(4./5.*8.27,4./5.*11.69)) # a4 portrait
    # mpl.rc('font', size=textsize)
    # mpl.rc('lines', linewidth=lwidth, color='black')
    # mpl.rc('axes', linewidth=alwidth, labelcolor='black')
    # mpl.rc('path', simplify=False) # do not remove

    # if (outtype == 'pdf'):
    #     print('Plot PDF ', pdffile)
    #     pdf_pages = PdfPages(pdffile)
    # elif (outtype == 'png'):
    #     print('Plot PNG ', pngbase)
    # else:
    #     print('Plot X')

    # ifig = 0
    # ifig += 1
    # iplot = 0
    # print('Plot - Fig ', ifig)
    # fig = plt.figure(ifig)

    # iplot += 1
    # sub    = fig.add_axes(position(nrow,ncol,iplot,hspace=hspace,vspace=vspace), polar=True)
    # clockplot(sub, si, sti, usetex=usetex, dolegend=True)

    # iplot += 1
    # sub    = fig.add_axes(position(nrow,ncol,iplot,hspace=hspace,vspace=vspace), polar=True)
    # clockplot(sub, si, sti, stierr, llybbox=1.00, usetex=usetex)

    # iplot += 1
    # sub    = fig.add_axes(position(nrow,ncol,iplot,hspace=hspace,vspace=vspace), polar=True)
    # clockplot(sub, [si,si[::-1]], [sti,sti[::-1]], [stierr,stierr[::-1]], llybbox=1.1, doabc=True, dolegend=True)

    # if (outtype == 'pdf'):
    #     pdf_pages.savefig(fig)
    #     plt.close(fig)
    # elif (outtype == 'png'):
    #     pngfile = pngbase+"{0:04d}".format(ifig)+".png"
    #     fig.savefig(pngfile, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)
    #     plt.close(fig)

    # if (outtype == 'pdf'):
    #     pdf_pages.close()
    # elif (outtype == 'png'):
    #     pass
    # else:
    #     plt.show()



#!/usr/bin/env python
import numpy as np
from pyjams.division import division
import pyjams.const as const


__all__ = ['interpol']


def interpol(xout, xin, yin):
    """
    One-dimensional linear interpolation on first dimension.

    If xin and yin are 1D arrays, this function wraps to numpy.interp.
    If yin is an ND array and xin and xout are 1D, then yin is interpolated in
    its first dimension.


    Definition
    ----------
    def interpol(xout, xin, yin):


    Input
    -----
    xout
        1D array of the x-coordinates of the interpolated values
        if yin is ND array.
        array-like of x-coordinates of the interpolated values
        if yin is 1D array.
    xin
        1D array of the x-coordinates of the data points, must be increasing.
    yin
        ND array of the y-coordinates of the data points,
        1st dim must have same length as xin.


    Output
    ------
    The interpolated values with shape (np.size(xin),)+yin.shape[1:].


    Examples
    --------
    >>> import numpy as np
    >>> xin  = np.arange(360, dtype=float)
    >>> yin  = np.sin(xin)
    >>> xout = np.arange(10)*10. + 0.5
    >>> soll = np.interp(xout, xin, yin)
    >>> yout = interpol(xout, xin, yin)
    >>> print(np.any(yout != soll))
    False

    >>> sout = (3,1)
    >>> yin2 = np.transpose(np.tile(yin,sout))
    >>> yout = interpol(xout, xin, yin2)
    >>> for i in range(3):
    ...    if np.any(yout[:,i] != soll):
    ...        print(True)

    >>> sout = (3,2,1)
    >>> yin3 = np.transpose(np.tile(yin,sout))
    >>> yout = interpol(xout, xin, yin3)
    >>> for i in range(3):
    ...    for j in range(2):
    ...        if np.any(yout[:,j,i] != soll):
    ...            print(True)


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python
    library, Department of Computational Hydrosystems, Helmholtz Centre for
    Environmental Research - UFZ, Leipzig, Germany.

    Copyright (c) 2012-2021 Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.


    History
    -------
    Written,  Matthias Cuntz, Jun 2012
    Modified, Matthias Cuntz, Feb 2013 - ported to Python 3
              Matthias Cuntz, Apr 2014 - assert
              Matthias Cuntz, Nov 2016 - const.tiny -> const.eps
              Matthias Cuntz, Sep 2021 - code refactoring
    """
    #
    # If yin 1D-array then call immediately np.interp without check
    # If no np.interp wanted uncomment next line
    # isone = False
    if np.ndim(yin) == 1:
        # If no np.interp wanted comment next line
        # and uncomment the following two lines
        return np.interp(xout, xin, yin)
        # isone = True
        # yin = yin[:,np.newaxis]
    #
    # Check input
    assert np.ndim(xin)  == 1, "x input values not 1D array"
    assert np.ndim(xout) <= 1, "x output values not scalar or 1D array"
    #
    # Subscripts
    eps  = const.eps
    s    = np.minimum(np.maximum(np.searchsorted(xin, xout)-1, 0),
                      xin.size-2)  # Subscript intervals
    # Distances
    ums1 = xout-xin[s]  # distance from point before
    ums2 = xin[s+1]-xin[s]
    ums  = division(ums1, ums2, 0.)
    ums  = np.where((np.abs(ums1) < eps) | (np.abs(ums2) < eps),
                    0., ums)  # for numerical stability
    # Blow to output shape
    sout = yin.shape[1:][::-1] + (1,)
    ums  = np.transpose(np.tile(ums, sout))

    # If no np.interp wanted comment next line and uncomment the following
    # five lines
    return yin[s, ...] + ums*(yin[s+1, ...] - yin[s, ...])
    # yout = yin[s,...] + ums*(yin[s+1,...]-yin[s,...])
    # if isone:
    #     return np.squeeze(yout)
    # else:
    #     return yout


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)
    # xin  = np.arange(360, dtype=float)
    # yin  = np.sin(xin)
    # xout = np.arange(10)*10. + 0.5
    # print np.interp(xout, xin, yin)
    # print interpol(xout, xin, yin)

    # sout = (10,1)
    # yin2 = np.transpose(np.tile(yin,sout))
    # yout = interpol(xout, xin, yin2)
    # print yout[:,0]
    # print yout[:,5]

    # sout = (3,2,1)
    # yin3 = np.transpose(np.tile(yin,sout))
    # yout = interpol(xout, xin, yin3)
    # print yout[:,0,0]
    # print yout[:,1,2]



"""
Provides physical, mathematical, computational, isotope, and material
constants, such as `Pi = 3.141592653589793238462643383279502884197`.

:copyright: Copyright 2012-2021 Matthias Cuntz, see AUTHORS.md for details.
:license: MIT License, see LICENSE for details.

Subpackages
===========
.. autosummary::
   const
"""
from .const import Pi, Pi2, Pi3, TwoPi, pi, pi2, pi3, Twopi, Sqrt2, sqrt2
from .const import gravity, T0, P0, T25, sigma, R, Rair, Rh2o
from .const import Na, kB, REarth
from .const import mmol_co2, molmass_co2, mmol_h2o, molmass_h2o
from .const import mmol_air, molmass_air
from .const import density_quartz, cheat_quartz, cheat_water, cheat_air
from .const import latentheat_vaporization
from .const import R13VPDB, R18VSMOW, R2VSMOW
from .const import tiny, huge, eps


__all__ = [
    'Pi', 'Pi2', 'Pi3', 'TwoPi', 'pi', 'pi2', 'pi3', 'Twopi', 'Sqrt2', 'sqrt2',
    'gravity', 'T0', 'P0', 'T25', 'sigma', 'R', 'Rair', 'Rh2o',
    'Na', 'kB', 'REarth', 'mmol_co2', 'molmass_co2', 'mmol_h2o', 'molmass_h2o',
    'mmol_air', 'molmass_air',
    'density_quartz', 'cheat_quartz', 'cheat_water', 'cheat_air',
    'latentheat_vaporization',
    'R13VPDB', 'R18VSMOW', 'R2VSMOW',
    'tiny', 'huge', 'eps']



#!/usr/bin/env python
"""
This is the unittest for gridcellarea module.

python -m unittest -v tests/test_gridcellarea.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_gridcellarea.py

"""
import unittest


def _flatten(itr):
    import numpy as np
    fitr = np.array(itr).flatten()
    if len(fitr) == 0:
        return list(fitr)
    else:
        if isinstance(fitr[0], str):
            return [ i for i in fitr ]
        else:
            return [ i if np.isfinite(i) else np.finfo(float).max
                     for i in fitr ]


class TestGridcellarea(unittest.TestCase):
    """
    Tests for gridcellarea.py
    """

    def test_gridcellarea(self):
        import numpy as np
        import pandas as pd
        from pyjams import gridcellarea

        lat = [0., 2.5, 5.0]
        lon = [0., 3.75, 7.5]

        rearth = 6371009.
        fsoll = [[1.15906555e+11, 1.15906555e+11, 1.15906555e+11],
                 [1.15796237e+11, 1.15796237e+11, 1.15796237e+11],
                 [1.15465495e+11, 1.15465495e+11, 1.15465495e+11]]

        rearth1 = 6371000.
        fsoll1 = [[1.15906227e+11, 1.15906227e+11, 1.15906227e+11],
                  [1.15795910e+11, 1.15795910e+11, 1.15795910e+11],
                  [1.15465169e+11, 1.15465169e+11, 1.15465169e+11]]

        # descending latitudes
        dlat = [0., -2.5, -5.0]

        # meridian within longitudes
        lon360 = [360., 3.75, 7.5]
        # dateline within longitudes
        lon180 = [180., -180.+3.75, -180.+7.5]

        # list
        fout  = gridcellarea(lat, lon)
        assert isinstance(fout, np.ndarray)
        self.assertEqual(_flatten(np.around(fout, -3)), _flatten(fsoll))

        # tuple, list
        fout  = gridcellarea(tuple(lat), lon)
        assert isinstance(fout, np.ndarray)
        self.assertEqual(_flatten(np.around(fout, -3)), _flatten(fsoll))

        # 2 tuple
        fout  = gridcellarea(tuple(lat), tuple(lon))
        assert isinstance(fout, np.ndarray)
        self.assertEqual(_flatten(np.around(fout, -3)), _flatten(fsoll))

        # array, list
        fout  = gridcellarea(np.array(lat), lon)
        assert isinstance(fout, np.ndarray)
        self.assertEqual(_flatten(np.around(fout, -3)), _flatten(fsoll))

        # array, pandas
        fout  = gridcellarea(np.array(lat), pd.Series(lon))
        assert isinstance(fout, np.ndarray)
        self.assertEqual(_flatten(np.around(fout, -3)), _flatten(fsoll))

        # 2 array
        fout  = gridcellarea(np.array(lat), np.array(lon))
        assert isinstance(fout, np.ndarray)
        self.assertEqual(_flatten(np.around(fout, -3)), _flatten(fsoll))

        # 2 pandas
        fout  = gridcellarea(pd.Series(lat), pd.Series(lon))
        assert isinstance(fout, np.ndarray)
        self.assertEqual(_flatten(np.around(fout, -3)), _flatten(fsoll))

        # rearth
        fout  = gridcellarea(lat, lon, rearth=rearth)
        assert isinstance(fout, np.ndarray)
        self.assertEqual(_flatten(np.around(fout, -3)), _flatten(fsoll))

        # rearth classic
        fout  = gridcellarea(lat, lon, rearth=rearth1)
        assert isinstance(fout, np.ndarray)
        self.assertEqual(_flatten(np.around(fout, -3)), _flatten(fsoll1))

        # globe
        fout  = gridcellarea(lat, lon, globe=True)
        fsoll2 = [[3.79774834e+12, 3.79774834e+12, 3.79774834e+12],
                  [1.15796240e+11, 1.15796240e+11, 1.15796240e+11],
                  [3.61823239e+12, 3.61823239e+12, 3.61823239e+12]]
        assert isinstance(fout, np.ndarray)
        self.assertEqual(_flatten(np.around(fout, -4)), _flatten(fsoll2))

        # descending lats
        fout  = gridcellarea(dlat, lon, globe=True)
        assert isinstance(fout, np.ndarray)
        self.assertEqual(_flatten(np.around(fout, -4)), _flatten(fsoll2))

        # meridian in lon
        fout  = gridcellarea(lat, lon360)
        assert isinstance(fout, np.ndarray)
        self.assertEqual(_flatten(np.around(fout, -3)), _flatten(fsoll))

        # date line in lon
        fout  = gridcellarea(lat, lon180)
        assert isinstance(fout, np.ndarray)
        self.assertEqual(_flatten(np.around(fout, -3)), _flatten(fsoll))

        # errors
        # lat > 90
        lat1 = [0., 2.5, 95.0]
        self.assertRaises(AssertionError, gridcellarea, lat1, lon)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
    Register colourmaps with Matplotlib or return RGB values.
    Colour maps include Cynthia Brewer's colour maps,
    maps from the Geography department of Oregon State University,
    and maps from PyNGL, the Python port of the NCAR Command Language (NCL).

    Provides also a routine to plot the colormaps.


    Definition
    ----------
    Register colormap with Matplotlib and get colormap handle
        def get_brewer(cname=None, names=False, rgb=False, rgb256=False, reverse=False, grey=False, gray=False):
    Register colormap with Matplotlib
        def register_brewer(cname='all', reverse=False, grey=False, gray=False):
    Print available colormap names
        def print_brewer(names='all'):
    Plot all colormaps in png files
       def plot_brewer(pngbase='brewer_colors_',reverse=False, grey=False, gray=False):


    Input
    -----
    None


    Optional Input
    --------------
    get_brewer
        cname      Name of colormap
        names      Return list of colormap names
                   if 'sequential': return sequential colormap names
                   if 'diverging': return diverging colormap names
                   if 'qualitative': return qualitative colormap names
                   if 'osu': return Oregon State University colormap names
                   if 'ncl_large': return ncl large colormap names with more than 50 colors
                   if 'ncl_small': return ncl small colormap names with more than 50 colors
                   if 'ncl_meteo_swiss': return ncl small colormap names from Meteo Swiss
                   if 'mma': return Mathematica colour maps
                   else: return sequential+diverging+qualitative+osu+ncl_large+ncl_small+ncl_meteo_swiss+mma colormap names
        rgb        if True: return RGB value tuple between 0 and 1
        rgb256     if True: return RGB value tuple between 0 and 255
        reverse    if True: reverse colormaps
        grey       if True: return grey equivalent
        gray       Same as grey
    plot_brewer
        pngbase    Name of png file bases. png files will be named pngbase001, pngbase002, ...
                   (default: 'brewer_colors_')
        reverse    if True: reverse colormaps
        grey       if True: return grey equivalent
        gray       Same as grey
    print_brewer
        names      Print colormap names (default: 'all')
                   if 'sequential': print sequential colormap names
                   if 'diverging': print diverging colormap names
                   if 'qualitative': print qualitative colormap names
                   if 'osu': print Oregon State University colormap names
                   if 'ncl_large': print ncl colormap names with more than 50 colors
                   if 'ncl_small': print ncl colormap names with less than 50 colors
                   if 'ncl_meteo_swiss': print ncl small colormap names from Meteo Swiss
                   if 'mma': return Mathematica colour maps
                   else: print sequential+diverging+qualitative+osu+ncl_large+ncl_small+ncl_meteo_swiss+mma colormap names
    register_brewer
        cname      Colormap to register colormaps (default: 'all')
                   if 'all': registers all sequential+diverging+qualitative+osu+ncl_large+ncl_small+ncl_meteo_swiss+mma colormaps
        reverse    if True: reverse colormaps
        grey       if True: return grey equivalent
        gray       Same as grey


    Output
    ------
    matplotlib colormap instance or tuple with RGB values


    References
    ----------
    Colourmaps by Cynthia A. Brewer, Geography, Pennsylvania State University
      http://colorbrewer2.org
    Values copied from
      http://www.personal.psu.edu/cab38/ColorBrewer/ColorBrewer_all_schemes_RGBonly4_withPalette_and_Macro.xls
    Colourmaps from Geography, University of Oregon
      http://geography.uoregon.edu/datagraphics/color_scales.htm
    Colourmaps from PyNGL, the Python interface to the NCAR Command Language (NCL)
      http://www.pyngl.ucar.edu/Graphics/color_table_gallery.shtml
    

    Examples
    --------
    # Print
    # print_brewer('diverging')

    # Plot
    # plot_brewer('brew_')

    # Get Names
    # names = get_brewer(names='sequential')

    # Register colour map and get colour map handle
    # cc = get_brewer('RdYlBu11')
    # plt.pcolormesh(np.outer(np.arange(cc.N), np.ones(cc.N)), cmap=cc)

    # Get RGB colours of colour map
    # cc = get_brewer('blues4', rgb=True)
    # mark1 = sub.plot(x, y)
    # plt.setp(mark1, linestyle='None', marker='o', markeredgecolor=cc[0], markerfacecolor='None')

    # Register and use colour map
    # from scipy import misc
    # lena = misc.lena()
    # plt.imshow(l, cmap=mpl.cm.rainbow)
    # register_brewer('ncl_meteo_swiss')
    # plt.imshow(l, cmap=mpl.cm.get_cmap('hotcold_18lev'))
    # plt.imshow(l, cmap=jams.get_brewer('hotcold_18lev'))


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2012-2014 Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  MC, Sep 2012
    Modified, MC, Feb 2013 - ported to Python 3
              MC, Jun 2013 - include Uni Oregon colours
              ST, Mar 2014 - include NCL color maps; define_brewer -> register_brewer
              MC, MAr 2014 - colour maps in extra file brewer.cmaps
              JM, Sep 2014 - color maps of Mathematica
"""
# Define colormaps
# from jams.find_in_path import find_in_path
# cmapfile = find_in_path('brewer.cmaps') # in jams_python
import os
cmapfile = os.path.join(os.path.dirname(__file__), 'brewer.cmaps') # in jams_python/jams
exec(compile(open(cmapfile).read(), cmapfile, 'exec'))

sequential_maps = ['ylgn3','ylgn4','ylgn5','ylgn6','ylgn7','ylgn8','ylgn9','ylgnbu3','ylgnbu4',
                   'ylgnbu5','ylgnbu6','ylgnbu7','ylgnbu8','ylgnbu9','gnbu3','gnbu4','gnbu5',
                   'gnbu6','gnbu7','gnbu8','gnbu9','bugn3','bugn4','bugn5','bugn6','bugn7','bugn8',
                   'bugn9','pubugn3','pubugn4','pubugn5','pubugn6','pubugn7','pubugn8','pubugn9',
                   'pubu3','pubu4','pubu5','pubu6','pubu7','pubu8','pubu9','bupu3','bupu4','bupu5',
                   'bupu6','bupu7','bupu8','bupu9','rdpu3','rdpu4','rdpu5','rdpu6','rdpu7','rdpu8',
                   'rdpu9','purd3','purd4','purd5','purd6','purd7','purd8','purd9','orrd3','orrd4',
                   'orrd5','orrd6','orrd7','orrd8','orrd9','ylorrd3','ylorrd4','ylorrd5','ylorrd6',
                   'ylorrd7','ylorrd8','ylorrd9','ylorbr3','ylorbr4','ylorbr5','ylorbr6','ylorbr7',
                   'ylorbr8','ylorbr9','purples3','purples4','purples5','purples6','purples7','purples8',
                   'purples9','blues3','blues4','blues5','blues6','blues7','blues8','blues9','greens3',
                   'greens4','greens5','greens6','greens7','greens8','greens9','oranges3','oranges4',
                   'oranges5','oranges6','oranges7','oranges8','oranges9','reds3','reds4','reds5','reds6',
                   'reds7','reds8','reds9','greys3','greys4','greys5','greys6','greys7','greys8','greys9']

diverging_maps = ['puor3','puor4','puor5','puor6','puor7','puor8','puor9','puor10','puor11','brbg3',
                  'brbg4','brbg5','brbg6','brbg7','brbg8','brbg9','brbg10','brbg11','prgn3','prgn4',
                  'prgn5','prgn6','prgn7','prgn8','prgn9','prgn10','prgn11','piyg3','piyg4','piyg5',
                  'piyg6','piyg7','piyg8','piyg9','piyg10','piyg11','rdbu3','rdbu4','rdbu5','rdbu6',
                  'rdbu7','rdbu8','rdbu9','rdbu10','rdbu11','rdgy3','rdgy4','rdgy5','rdgy6','rdgy7',
                  'rdgy8','rdgy9','rdgy10','rdgy11','rdylbu3','rdylbu4','rdylbu5','rdylbu6','rdylbu7',
                  'rdylbu8','rdylbu9','rdylbu10','rdylbu11','spectral3','spectral4','spectral5',
                  'spectral6','spectral7','spectral8','spectral9','spectral10','spectral11',
                  'rdylgn3','rdylgn4','rdylgn5','rdylgn6','rdylgn7','rdylgn8','rdylgn9','rdylgn10',
                  'rdylgn11']

qualitative_maps = ['set33','set34','set35','set36','set37','set38','set39','set310','set311','set312',
                    'pastel13','pastel14','pastel15','pastel16','pastel17','pastel18','pastel19','set13',
                    'set14','set15','set16','set17','set18','set19','pastel23','pastel24','pastel25',
                    'pastel26','pastel27','pastel28','set23','set24','set25','set26','set27','set28',
                    'dark23','dark24','dark25','dark26','dark27','dark28','paired3','paired4','paired5',
                    'paired6','paired7','paired8','paired9','paired10','paired11','paired12','accent3',
                    'accent4','accent5','accent6','accent7','accent8']

osu_maps = ['bu7', 'bu10', 'step5',
            'brbu10', 'brbu12', 'budor12', 'budor18', 'drdbu12', 'drdbu18', 'bugn14', 'bugr8',
            'buor8', 'buor10', 'buor12', 'buor14', 'rdylbu11b', 'gnmg16',
            'cat12']

ncl_large_maps = ['BkBlAqGrYeOrReViWh200', 'BlAqGrYeOrRe', 'BlAqGrYeOrReVi200', 
                  'BlGrYeOrReVi200', 'BlRe', 'BlueRed', 'BlueRedGray', 'BlueWhiteOrangeRed',
                  'BlueYellowRed', 'BlWhRe', 'detail', 'example', 'extrema', 'GrayWhiteGray',
                  'GreenYellow', 'helix', 'helix1', 'hotres', 'ncview_default',
                  'OceanLakeLandSnow', 'psgcap', 'rainbow', 'rainbowwhitegray', 
                  'rainbowwhite', 'rainbowgray', 'testcmap', 'tbr_240_300',
                  'tbr_stdev_0_30', 'tbr_var_0_500', 'tbrAvg1', 'tbrStd1', 'tbrVar1',
                  'temp1', 'thelix', 'uniform', 'ViBlGrWhYeOrRe', 'wh_bl_gr_ye_re', 
                  'WhBlGrYeRe', 'WhBlReWh', 'WhViBlGrYeOrRe', 'WhViBlGrYeOrReWh', 
                  'WhiteBlue', 'WhiteBlueGreenYellowRed', 'WhiteGreen', 
                  'WhiteYellowOrangeRed', 'wxpEnIR', 'gauss', 'saw']

ncl_small_maps = ['amwg', 'amwg_blueyellowred', 'BlueDarkRed18', 
                  'BlueDarkOrange18', 'BlueGreen14', 'BrownBlue12', 
                  'Cat12', 'cosam12', 'cosam', 'cyclic', 'default',
                  'GreenMagenta16', 'gscyclic', 'gsdtol', 'gsltod',
                  'gui_default', 'hlu_default', 'hotcold_18lev',
                  'hotcolr_19lev', 'mch_default', 'nrl_sirkes', 
                  'perc2_9lev', 'percent_11lev', 'posneg_1', 'posneg_2',
                  'prcp_1', 'prcp_2', 'prcp_3', 'precip_11lev',
                  'precip_diff_12lev', 'precip_diff_1lev', 'precip2_15lev',
                  'precip2_17lev', 'precip3_16lev', 'precip4_11lev',
                  'precip4_diff_19lev', 'radar', 'radar_1', 'rh_19lev',
                  'so4_21', 'so4_23', 'spread_15lev', 'StepSeq25', 'sunshine_9lev',
                  'sunshine_diff_12lev', 'temp_19lev', 'temp_diff_18lev',
                  'temp_diff_1lev', 'topo_15lev', 'wgne15', 'wind_17lev']

ncl_meteo_swiss_maps = ['hotcold_18lev', 'hotcolr_19lev', 'mch_default', 'perc2_9lev',
                        'percent_11lev', 'precip2_15lev', 'precip2_17lev', 'precip3_16lev',
                        'precip4_11lev', 'precip4_diff_19lev', 'precip_11lev', 'precip_diff_12lev',
                        'precip_diff_1lev', 'rh_19lev', 'spread_15lev', 'sunshine_9lev', 'sunshine_diff_12lev',
                        't2m_29lev', 'temp_19lev', 'temp_diff_18lev', 'temp_diff_1lev', 'topo_15lev',
                        'wind_17lev']

mma_maps = ['dark_rainbow_6', 'dark_rainbow_8', 'dark_rainbow_12', 'dark_rainbow_16', 'dark_rainbow_256',
            'starry_night_colors_8','starry_night_colors_16','starry_night_colors_256',
            'gray_yellow_tones_8', 'gray_yellow_tones_16', 'gray_yellow_tones_256']

all_maps = sequential_maps + diverging_maps + qualitative_maps + osu_maps + ncl_large_maps + ncl_small_maps + ncl_meteo_swiss_maps + mma_maps

def capitalise(cname):
    ''' Search for colour map if not the right capitalisation is given.
        Ex.:     cmaps = [ capitalise(cc) for cc in cmaps ]
    '''
    if cname in all_maps:
        return cname
    else:
        lcmaps = [ c.lower() for c in all_maps ]
        if cname.lower() in lcmaps:
            return all_maps[lcmaps.index(cname.lower())]
        else:
            raise ValueError('Color map name not known: '+cname)            

def register_brewer(cname='all', reverse=False, grey=False, gray=False):
    '''Register colourmap(s) with Matplotlib
       Ex.: register_brewer('ygn3')
    '''
    import matplotlib.colors as col
    import matplotlib.cm as cm
    if cname.lower() == 'all':
        cmaps = all_maps
    elif cname.lower() == 'sequential':
        cmaps = sequential_maps
    elif cname.lower() == 'diverging':
        cmaps = diverging_maps
    elif cname.lower() == 'qualitative':
        cmaps = qualitative_maps
    elif cname.lower() == 'osu':
        cmaps = osu_maps
    elif cname.lower() == 'ncl_large':
        cmaps = ncl_large_maps
    elif cname.lower() == 'ncl_small':
        cmaps = ncl_small_maps
    elif cname.lower() == 'ncl_meteo_swiss':
        cmaps = ncl_meteo_swiss_maps
    elif cname.lower() == 'mma':
        cmaps = mma_maps
    else:
        cmaps = [cname]
    cmaps = [ capitalise(cc) for cc in cmaps ]
    for i in cmaps:
        d = {}
        if i in ncl_large_maps + ncl_small_maps + ncl_meteo_swiss_maps + mma_maps:
            exec('cpool = '+i, globals(), d)
        else:
            exec('cpool = [ tuple([k/255. for k in j]) for j in '+i+' ]', globals(), d)
        cpool = d['cpool']
        if reverse:
            cpool = cpool[::-1]
        if grey | gray:
            for j in range(len(cpool)):
                isgray = 0.2125 * cpool[j][0] + 0.7154 * cpool[j][1] + 0.072* cpool[j][2]
                cpool[j] = (isgray,isgray,isgray)
        cmap = col.ListedColormap(cpool, i)
        cm.register_cmap(cmap=cmap)


def print_brewer(names='all'):
    """
        Print names of known colour maps.

        Examples
        --------
        >>> print_brewer('qualitative')
        ['set33', 'set34', 'set35', 'set36', 'set37', 'set38', 'set39', 'set310', 'set311', 'set312', 'pastel13', 'pastel14', 'pastel15', 'pastel16', 'pastel17', 'pastel18', 'pastel19', 'set13', 'set14', 'set15', 'set16', 'set17', 'set18', 'set19', 'pastel23', 'pastel24', 'pastel25', 'pastel26', 'pastel27', 'pastel28', 'set23', 'set24', 'set25', 'set26', 'set27', 'set28', 'dark23', 'dark24', 'dark25', 'dark26', 'dark27', 'dark28', 'paired3', 'paired4', 'paired5', 'paired6', 'paired7', 'paired8', 'paired9', 'paired10', 'paired11', 'paired12', 'accent3', 'accent4', 'accent5', 'accent6', 'accent7', 'accent8']
    """
    if names.lower() == 'sequential':
        print(sequential_maps)
    elif names.lower() == 'diverging':
        print(diverging_maps)
    elif names.lower() == 'qualitative':
        print(qualitative_maps)
    elif names.lower() == 'osu':
        print(osu_maps)
    elif names.lower() == 'ncl_large':
        print(ncl_large_maps)
    elif names.lower() == 'ncl_small':
        print(ncl_small_maps)
    elif names.lower() == 'ncl_meteo_swiss':
        print(ncl_meteo_swiss_maps)
    elif names.lower() == 'mma':
        print(mma_maps)
    else:
        print('Sequential color maps')
        print(sequential_maps)
        print('')
        print('Diverging color maps')
        print(diverging_maps)
        print('')
        print('Qualitative color maps')
        print(qualitative_maps)
        print('')
        print('Oregon State University color maps')
        print(osu_maps)
        print('')
        print('NCL large color maps')
        print(ncl_large_maps)
        print('')
        print('NCL small color maps')
        print(ncl_small_maps)
        print('')
        print('NCL meteo swiss color maps')
        print(ncl_meteo_swiss_maps)
        print('')
        print('Mathematica color maps')
        print(mma_maps)

def get_brewer(cname=None, names=False, rgb=False, rgb256=False, reverse=False, grey=False, gray=False):
    """
        Get colour map either as registered handle or as RGB values (0-255 or 0-1).

        Examples
        --------
        >>> import numpy as np
        >>> from autostring import astr
        >>> cc = get_brewer('blues4',rgb=True)
        >>> print(astr(np.array(cc[0]), 4))
        ['0.9373' '0.9529' '1.0000']

        >>> cc = get_brewer('Blues4',rgb256=True)
        >>> print(cc[0])
        (239, 243, 255)

        >>> cc = get_brewer('bLuEs4',rgb256=True,reverse=True)
        >>> print(cc[-1])
        (239, 243, 255)
        >>> print(cc[0])
        (33, 113, 181)

        >>> cc = get_brewer('blues4',rgb256=True,grey=True)
        >>> print(astr(np.array(cc[0]), 4))
        ['242.9897' '242.9897' '242.9897']
    """
    if names:
        if names.lower() == 'sequential':
            return sequential_maps
        elif names.lower() == 'diverging':
            return diverging_maps
        elif names.lower() == 'qualitative':
            return qualitative_maps
        elif names.lower() == 'osu':
            return osu_maps
        elif names.lower() == 'ncl_large':
            return ncl_large_maps
        elif names.lower() == 'ncl_small':
            return ncl_small_maps
        elif names.lower() == 'ncl_meteo_swiss':
            return ncl_meteo_swiss_maps
        elif names.lower() == 'mma':
            return mma_maps
        else:
            cmaps = all_maps
            return cmaps
    else:
        cname = capitalise(cname)
        if rgb256:
            d = {}
            if cname in ncl_large_maps + ncl_small_maps + ncl_meteo_swiss_maps + mma_maps:
                exec('cpool = [ tuple([k*255. for k in j]) for j in '+cname+' ]', globals(), d)
            else:
                exec('cpool = '+cname, globals(), d)
            cpool = d['cpool']
            if reverse:
                cpool = cpool[::-1]
            if grey | gray:
                for j in range(len(cpool)):
                    isgray = 0.2125 * cpool[j][0] + 0.7154 * cpool[j][1] + 0.072* cpool[j][2]
                    cpool[j] = (isgray,isgray,isgray)
            return cpool
        # get colour tuple in 0-1
        elif rgb:
            d = {}
            if cname in ncl_large_maps + ncl_small_maps + ncl_meteo_swiss_maps + mma_maps:
                exec('cpool = '+cname, globals(), d)
            else:
                exec('cpool = [ tuple([k/255. for k in j]) for j in '+cname+' ]', globals(), d)
            cpool = d['cpool']
            if reverse:
                cpool = cpool[::-1]
            if grey | gray:
                for j in range(len(cpool)):
                    isgray = 0.2125 * cpool[j][0] + 0.7154 * cpool[j][1] + 0.072* cpool[j][2]
                    cpool[j] = (isgray,isgray,isgray)
            return cpool
        # register colour map with matplotlib
        else:
            import matplotlib.cm as cm
            register_brewer(cname,reverse=reverse, grey=grey, gray=gray)
            return cm.get_cmap(cname)


def plot_brewer(pngbase='brewer_colors_',reverse=False, grey=False, gray=False):
    '''Plot all colour bars in PNG files.
       Ex.: plot_brewer('brewer_', reverse=True)
    '''
    import numpy as np
    from position import position
    outtype = 'pdf'
    nrow       = 5           # # of rows of subplots per figure
    ncol       = 14          # # of columns of subplots per figure
    hspace     = 0.01        # x-space between subplots
    wspace     = 0.08        # y-space between subplots
    textsize   = 10          # standard text size
    lwidth     = 1.0         # axis line width
    alwidth    = 1.0         # axis line width
    dpi         = 300
    transparent = False
    bbox_inches = 'tight'
    pad_inches  = 0
    import matplotlib as mpl
    mpl.use('Agg') # set directly after import matplotlib
    import matplotlib.pyplot as plt
    mpl.rc('figure', figsize=(8.27,11.69)) # a4 portrait
    #mpl.rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
    mpl.rc('font',**{'family':'serif','serif':['times']})
    mpl.rc('text.latex', unicode=True)
    mpl.rc('savefig', dpi=dpi, format='png')
    mpl.rc('font', size=textsize)
    mpl.rc('lines', linewidth=lwidth, color='black')
    mpl.rc('axes', linewidth=alwidth, labelcolor='black')
    mpl.rc('path', simplify=False) # do not remove

    figsize = mpl.rcParams['figure.figsize']
    ifig = 0

    nmaps = len(sequential_maps)
    npage = nrow*ncol
    zaehl = 0
    for kk in sequential_maps:
        if zaehl % npage == 0:
            if zaehl != 0:
                pngfile = pngbase+"{0:04d}".format(ifig)+".png"
                fig.savefig(pngfile, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)
                plt.close(fig)
            ifig += 1
            iplot = 0
            print('Page ', ifig)
            fig = plt.figure(ifig)
        iplot += 1
        sub = fig.add_axes(position(nrow,ncol,iplot,hspace=hspace,wspace=wspace))
        sub.axis('off')
        if iplot == 1:
            fig.text(0.5, 0.97, "Sequential", ha="center", size=12)
        cc = get_brewer(kk, reverse=reverse, grey=grey, gray=gray)
        plt.pcolormesh(np.outer(np.arange(cc.N),np.ones(cc.N)),cmap=cc)
        plt.setp(sub, title=kk)
        plt.setp(sub.title,fontsize=10, rotation=90, verticalalignment='bottom')
        zaehl += 1
    pngfile = pngbase+"{0:04d}".format(ifig)+".png"
    fig.savefig(pngfile, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)
    plt.close(fig)

    nmaps = len(diverging_maps)
    npage = nrow*ncol
    zaehl = 0
    for kk in diverging_maps:
        if zaehl % npage == 0:
            if zaehl != 0:
                pngfile = pngbase+"{0:04d}".format(ifig)+".png"
                fig.savefig(pngfile, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)
                plt.close(fig)
            ifig += 1
            iplot = 0
            print('Page ', ifig)
            fig = plt.figure(ifig)
        iplot += 1
        sub = fig.add_axes(position(nrow,ncol,iplot,hspace=hspace,wspace=wspace))
        sub.axis('off')
        if iplot == 1:
            fig.text(0.5, 0.97, "Diverging", ha="center", size=12)
        cc = get_brewer(kk,reverse=reverse, grey=grey, gray=gray)
        plt.pcolormesh(np.outer(np.arange(cc.N),np.ones(cc.N)),cmap=cc)
        plt.setp(sub, title=kk)
        plt.setp(sub.title,fontsize=10, rotation=90, verticalalignment='bottom')
        zaehl += 1
    pngfile = pngbase+"{0:04d}".format(ifig)+".png"
    fig.savefig(pngfile, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)
    plt.close(fig)

    nmaps = len(qualitative_maps)
    npage = nrow*ncol
    zaehl = 0
    for kk in qualitative_maps:
        if zaehl % npage == 0:
            if zaehl != 0:
                pngfile = pngbase+"{0:04d}".format(ifig)+".png"
                fig.savefig(pngfile, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)
                plt.close(fig)
            ifig += 1
            iplot = 0
            print('Page ', ifig)
            fig = plt.figure(ifig)
        iplot += 1
        sub = fig.add_axes(position(nrow,ncol,iplot,hspace=hspace,wspace=wspace))
        sub.axis('off')
        if iplot == 1:
            fig.text(0.5, 0.97, "Qualitative", ha="center", size=12)
        cc = get_brewer(kk,reverse=reverse, grey=grey, gray=gray)
        plt.pcolormesh(np.outer(np.arange(cc.N),np.ones(cc.N)),cmap=cc)
        plt.setp(sub, title=kk)
        plt.setp(sub.title,fontsize=10, rotation=90, verticalalignment='bottom')
        zaehl += 1
    pngfile = pngbase+"{0:04d}".format(ifig)+".png"
    fig.savefig(pngfile, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)
    plt.close(fig)

    nmaps = len(osu_maps)
    npage = nrow*ncol
    zaehl = 0
    for kk in osu_maps:
        if zaehl % npage == 0:
            if zaehl != 0:
                pngfile = pngbase+"{0:04d}".format(ifig)+".png"
                fig.savefig(pngfile, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)
                plt.close(fig)
            ifig += 1
            iplot = 0
            print('Page ', ifig)
            fig = plt.figure(ifig)
        iplot += 1
        sub = fig.add_axes(position(nrow,ncol,iplot,hspace=hspace,wspace=wspace))
        sub.axis('off')
        if iplot == 1:
            fig.text(0.5, 0.97, "Oregon State University (OSU)", ha="center", size=12)
        cc = get_brewer(kk,reverse=reverse, grey=grey, gray=gray)
        plt.pcolormesh(np.outer(np.arange(cc.N),np.ones(cc.N)),cmap=cc)
        plt.setp(sub, title=kk)
        plt.setp(sub.title,fontsize=10, rotation=90, verticalalignment='bottom')
        zaehl += 1
    pngfile = pngbase+"{0:04d}".format(ifig)+".png"
    fig.savefig(pngfile, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)
    plt.close(fig)

    nrow       = 1  # only one row for ncl large maps
    ncol       = 14
    nmaps = len(ncl_large_maps)
    npage = nrow*ncol
    zaehl = 0
    for kk in ncl_large_maps:
        if zaehl % npage == 0:
            if zaehl != 0:
                pngfile = pngbase+"{0:04d}".format(ifig)+".png"
                fig.savefig(pngfile, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)
                plt.close(fig)
            ifig += 1
            iplot = 0
            print('Page ', ifig)
            fig = plt.figure(ifig)
        iplot += 1
        sub = fig.add_axes(position(nrow,ncol,iplot,hspace=hspace,wspace=wspace,top=0.85))
        sub.axis('off')
        if iplot == 1:
            fig.text(0.5, 0.97, "ncl_large_maps", ha="center", size=12)
        cc = get_brewer(kk,reverse=reverse, grey=grey, gray=gray)
        plt.pcolor(np.outer(np.arange(cc.N),np.ones(cc.N)),cmap=cc,linewidth=0,edgecolor='None')
        plt.setp(sub, title=kk)
        plt.setp(sub.title,fontsize=10, rotation=90, verticalalignment='bottom')
        zaehl += 1
    pngfile = pngbase+"{0:04d}".format(ifig)+".png"
    fig.savefig(pngfile, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)
    plt.close(fig)
    
    nrow       = 2 # set to two for ncl small maps
    ncol       = 14
    nmaps = len(ncl_small_maps)
    npage = nrow*ncol
    zaehl = 0
    for kk in ncl_small_maps:
        if zaehl % npage == 0:
            if zaehl != 0:
                pngfile = pngbase+"{0:04d}".format(ifig)+".png"
                fig.savefig(pngfile, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)
                plt.close(fig)
            ifig += 1
            iplot = 0
            print('Page ', ifig)
            fig = plt.figure(ifig)
        iplot += 1
        sub = fig.add_axes(position(nrow,ncol,iplot,hspace=hspace,wspace=wspace,top=0.85))
        sub.axis('off')
        if iplot == 1:
            fig.text(0.5, 0.97, "ncl_small_maps", ha="center", size=12)
        cc = get_brewer(kk,reverse=reverse, grey=grey, gray=gray)
        plt.pcolor(np.outer(np.arange(cc.N),np.ones(cc.N)),cmap=cc,linewidth=0,edgecolor='None')
        plt.setp(sub, title=kk)
        plt.setp(sub.title,fontsize=10, rotation=90, verticalalignment='bottom')
        zaehl += 1
    pngfile = pngbase+"{0:04d}".format(ifig)+".png"
    fig.savefig(pngfile, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)
    plt.close(fig)
    
    nmaps = len(ncl_meteo_swiss_maps)
    npage = nrow*ncol
    zaehl = 0
    for kk in ncl_meteo_swiss_maps:
        if zaehl % npage == 0:
            if zaehl != 0:
                pngfile = pngbase+"{0:04d}".format(ifig)+".png"
                fig.savefig(pngfile, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)
                plt.close(fig)
            ifig += 1
            iplot = 0
            print('Page ', ifig)
            fig = plt.figure(ifig)
        iplot += 1
        sub = fig.add_axes(position(nrow,ncol,iplot,hspace=hspace,wspace=wspace,top=0.85))
        sub.axis('off')
        if iplot == 1:
            fig.text(0.5, 0.97, "ncl_meteo_swiss_maps", ha="center", size=12)
        cc = get_brewer(kk,reverse=reverse, grey=grey, gray=gray)
        plt.pcolor(np.outer(np.arange(cc.N),np.ones(cc.N)),cmap=cc,linewidth=0,edgecolor='None')
        plt.setp(sub, title=kk)
        plt.setp(sub.title,fontsize=10, rotation=90, verticalalignment='bottom')
        zaehl += 1
    pngfile = pngbase+"{0:04d}".format(ifig)+".png"
    fig.savefig(pngfile, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)
    plt.close(fig)

    nmaps = len(mma_maps)
    npage = nrow*ncol
    zaehl = 0
    for kk in mma_maps:
        if zaehl % npage == 0:
            if zaehl != 0:
                pngfile = pngbase+"{0:04d}".format(ifig)+".png"
                fig.savefig(pngfile, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)
                plt.close(fig)
            ifig += 1
            iplot = 0
            print('Page ', ifig)
            fig = plt.figure(ifig)
        iplot += 1
        sub = fig.add_axes(position(nrow,ncol,iplot,hspace=hspace,wspace=wspace,top=0.85))
        sub.axis('off')
        if iplot == 1:
            fig.text(0.5, 0.97, "mma_maps", ha="center", size=12)
        cc = get_brewer(kk,reverse=reverse, grey=grey, gray=gray)
        plt.pcolor(np.outer(np.arange(cc.N),np.ones(cc.N)),cmap=cc,linewidth=0,edgecolor='None')
        plt.setp(sub, title=kk)
        plt.setp(sub.title,fontsize=10, rotation=90, verticalalignment='bottom')
        zaehl += 1
    pngfile = pngbase+"{0:04d}".format(ifig)+".png"
    fig.savefig(pngfile, transparent=transparent, bbox_inches=bbox_inches, pad_inches=pad_inches)
    plt.close(fig)
    
if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # import numpy as np
    # from autostring import astr
    # cc = get_brewer('blues4',rgb=True)
    # print(astr(np.array(cc[0]), 4))
    # #    ['0.9373' '0.9529' '1.0000']

    # cc = get_brewer('Blues4',rgb256=True)
    # print(cc[0])
    # #    (239, 243, 255)

    # cc = get_brewer('blues4',rgb256=True,reverse=True)
    # print(cc[-1])
    # #    (239, 243, 255)
    # print(cc[0])
    # #    (33, 113, 181)

    # cc = get_brewer('blues4',rgb256=True,grey=True)
    # print(astr(np.array(cc[0]), 4))
    # #    ['242.9897' '242.9897' '242.9897']

    # plot_brewer('test_', reverse=True)
    # plot_brewer('test_gray_', gray=True)

    # # Register and use colour map
    # from scipy import misc
    # lena = misc.lena()
    # from matplotlib.pylab import *
    # imshow(lena, cmap=mpl.cm.rainbow)
    # show()
    # register_brewer('ncl_meteo_swiss')
    # imshow(lena, cmap=mpl.cm.get_cmap('hotcold_18lev'))
    # show()
    # imshow(lena, cmap=get_brewer('BlueRedGray'))
    # show()

    # # Register and use colour map
    # from scipy import misc
    # lena = misc.lena()
    # from matplotlib.pylab import *
    # imshow(lena, cmap=mpl.cm.rainbow)
    # show()
    # jams.register_brewer('mma')
    # # jams.print_brewer('mma')
    # imshow(lena, mpl.cm.get_cmap('dark_rainbow_16'))
    # show()



#!/usr/bin/env python
"""
Morris' Method of Elementary Effects.

Module includes optimised sampling of trajectories including optional groups as
well as calculation of the Morris measures mu, stddev and mu*.

Saltelli, A., Chan, K., & Scott, E. M. (2000). Sensitivity Analysis. Wiley
Series in Probability and Statistics, John Wiley & Sons, New York, 1-504,
pages 68ff

Provided functions are:
    morris_sampling - Sample trajectories in parameter space

    elementary_effects - Calculate Elementary Effects from model
    output on trajectories

Note that the functions morris_sampling and elementary_effects are wrappers for
the functions Optimized_Groups and Morris_Measure_Groups of F. Campolongo and
J. Cariboni ported to Python by Stijn Van Hoey.

This module was originally written by Stijn Van Hoey as a translation
from an original Matlab code of F. Campolongo and J. Cariboni, JRC -
IPSC Ispra, Varese, IT. It was adapted by Matthias Cuntz while at
Department of Computational Hydrosystems, Helmholtz Centre for
Environmental Research - UFZ, Leipzig, Germany, and continued while at
Institut National de Recherche pour l'Agriculture, l'Alimentation et
l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2012-2022 Stijn Van Hoey, Matthias Cuntz, see AUTHORS.rst
            for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following wrappers are provided

.. autosummary::
   morris_sampling
   elementary_effects

History
    * Written in Matlab by F. Campolongo, J. Cariboni,
      JRC - IPSC Ispra, Varese, IT.
      Last Update 15 November 2005 by J. Cariboni:
      http://sensitivity-analysis.jrc.ec.europa.eu/software/index.htm
      now at: https://ec.europa.eu/jrc/en/samo/simlab
    * Translated to Python in May 2012 by Stijn Van Hoey.
    * Adapted to Python 3, etc., Oct 2013, Matthias Cuntz
    * Went from exponential time increase with number of trajectories to linear
      increase by using in Optimised_Groups one call to cdist from
      scipy.spatial.distance and removed one loop in a loop over total number
      of trajectories.
      Several further little improvements on speed, Dec 2017, Matthias Cuntz
    * Allow single trajectories, Dec 2017, Matthias Cuntz
    * Catch degenerated case where lower bound==upper bound, return 0,
      Feb 2018, Matthias Cuntz
    * Use integer division operator // instead of / for trajectory length r,
      Jul 2018, Fabio Genaretti
    * Distance matrix is not done for all trajectories at once because of very
      large memory requirement, Aug 2018, Matthias Cuntz & Fabio Gennaretti
    * Changed to Sphinx docstring and numpydoc, Dec 2019, Matthias Cuntz
    * Distinguish iterable and array_like parameter types,
      Jan 2020, Matthias Cuntz
    * Remove np.matrix in Sampling_Function_2, called in Optimized_Groups to
      remove numpy deprecation warnings, Jan 2020, Matthias Cuntz
    * Plot diagnostic figures in png files if matplotlib installed, Feb 2020,
      Matthias Cuntz
    * Renamed file to morris_method.py, Feb 2020, Matthias Cuntz
    * Adjusted argument and keyword argument names to be consistent with
      pyeee, Feb 2020, Matthias Cuntz
    * Make number of final trajectories an argument instead of a keyword
      argument and sample default of 10*final trajectories,
      Feb 2020, Matthias Cuntz
    * Sample not only from uniform distribution but allow all distributions of
      scipy.stats, Mar 2020, Matthias Cuntz
    * More consistent docstrings, Jan 2022, Matthias Cuntz

"""
import numpy as np


__all__ = ['morris_sampling', 'elementary_effects']


def Sampling_Function_2(p, k, r, lb, ub, GroupMat=np.array([])):
    """
    Morris sampling function


    Definition
    ----------
    def  Sampling_Function_2(p, k, r, lb, ub, GroupMat=np.array([])):


    Input
    -----
    p : int
        number of intervals considered in [0,1]
    k : int
        number of factors examined (``sizea=k``).
        In case groups are chosen the number of factors is stored in
        ``NumFact`` and ``sizea`` becomes the number of created groups
        (``sizea=GroupMat.shape[1]``).
    r : int
        sample size
    lb : array-like of size (``sizea``,)
        Lower Bound for each factor in list or array
    ub : array-like of size (``sizea``,)
        Upper Bound for each factor in list or array


    Optional Input
    --------------
    GroupMat : ndarray with size (``NumFact``, ``GroupNumber``)
        Array which describes the chosen groups. (default: ``np.array([])``)
        Each column represents a group and its elements are set to 1
        in correspondence of the factors that belong to the fixed group. All
        the other elements are zero.


    Output
    ------
    list
        [``Outmatrix``, ``OutFact``]
        ``Outmatrix(sizeb*r, sizea)`` for the entire sample size computed
            ``In(i,j)`` matrices
        ``OutFact(sizea*r)` `for the entire sample size computed ``Fact(i,1)``
            vectors


    Notes
    -----
    Local Variables
        NumFact : number of factors examined in the case when groups are chosen
        GroupNumber : Number of groups (eventually 0)
        sizeb : sizea+1
        randmult(sizea) : vector of random +1 and -1
        perm_e(sizea) : vector of sizea random permutated indeces
        fact(sizea) : vector containing the factor varied within each traj
        DDo(sizea,sizea) : D* in Morris, 1991
        A(sizeb,sizea) : Jk+1,k in Morris, 1991
        B(sizeb,sizea) : B in Morris, 1991
        Po(sizea,sizea) : P* in Morris, 1991
        Bo(sizeb,sizea) : B* in Morris, 1991
        Ao(sizeb) : Jk+1 in Morris, 1991
        xo(sizea) : x* in Morris, 1991 (starting point for the trajectory)
        In(sizeb,sizea) : for each loop orientation matrix. It corresponds to
            a trajectory of k step in the parameter space and it provides a
            single elementary effect per factor
        Fact(sizea) : For each loop vector indicating which factor or group of
            factors has been changed in each step of the trajectory
        AuxMat(sizeb,sizea) : Delta*0.5*((2*B - A) * DD0 + A) in Morris, 1991.
            The AuxMat is used as in Morris design for single factor analysis,
            while it constitutes an intermediate step for the group analysis.

    Note: B0 is constructed as in Morris design when groups are not considered.

        When groups are considered the routine follows the following steps:

           1. Creation of P0 and DD0 matrices defined in Morris for the
              groups. This means that the dimensions of these 2 matrices are
              (GroupNumber,GroupNumber).

           2. Creation of AuxMat matrix with (GroupNumber+1,GroupNumber)
              elements.

           3. Definition of GroupB0 starting from AuxMat, GroupMat and P0.

           4. The final B0 for groups is obtained as [ones(sizeb,1)*x0' +
              GroupB0]. The P0 permutation is present in GroupB0 and it's
              not necessary to permute the matrix (ones(sizeb,1)*x0')
              because it's already randomly created.


    References
    ----------
    Saltelli, A., Chan, K., & Scott, E. M. (2000). Sensitivity Analysis.
        Wiley Series in Probability and Statistics, John Wiley & Sons,
        New York, 1-504. - on page 68ff


    History
    -------
    Written original Matlab code by F. Campolongo, J. Cariboni,
        JRC - IPSC Ispra, Varese, IT
        Last Update: 15 November 2005 by J.Cariboni
        http://sensitivity-analysis.jrc.ec.europa.eu/software/index.htm
        now at: https://ec.europa.eu/jrc/en/samo/simlab
    Modified, Stijn Van Hoey, May 2012 - ported to Python
              Matthias Cuntz, Oct 2013
                  - adapted to JAMS Python package and ported to Python 3
              Matthias Cuntz, Jan 2020 - remove np.matrix
    """
    # Parameters and initialisation of the output matrix
    sizea = k
    Delta = p / (2. * (p - 1.))
    NumFact = sizea
    if GroupMat.shape[0] == GroupMat.size:
        Groupnumber = 0
    else:
        Groupnumber = GroupMat.shape[1]  # size(GroupMat,2)
        sizea = GroupMat.shape[1]

    sizeb = sizea + 1

    Outmatrix = np.zeros(((sizea + 1) * r, NumFact))
    OutFact = np.zeros(((sizea + 1) * r, 1))
    # For each i generate a trajectory
    for i in range(r):
        Fact = np.zeros(sizea + 1)
        # Construct DD0
        DD0 = np.diagflat(np.sign(np.random.random(k) * 2 - 1))

        # Construct B (lower triangular)
        B = np.tri((sizeb), sizea, k=-1, dtype=int)

        # Construct A0, A
        A0 = np.ones((sizeb, 1))
        A  = np.ones((sizeb, NumFact))

        # Construct the permutation matrix P0. In each column of P0 one
        # randomly chosen element equals 1 while all the others equal zero. P0
        # tells the order in which order factors are changed in each Note that
        # P0 is then used reading it by rows.
        I = np.eye(sizea)
        P0 = I[:, np.random.permutation(sizea)]

        # When groups are present the random permutation is done only on B. The
        # effect is the same since the added part (A0*x0') is completely
        # random.
        if Groupnumber != 0:
            B = np.dot(B, np.dot(GroupMat, P0.T).T)

        # Compute AuxMat both for single factors and groups analysis. For
        # Single factors analysis AuxMat is added to (A0*X0) and then
        # permutated through P0. When groups are active AuxMat is used to build
        # GroupB0. AuxMat is created considering DD0. If the element on DD0
        # diagonal is 1 then AuxMat will start with zero and add Delta. If the
        # element on DD0 diagonal is -1 then DD0 will start Delta and goes to
        # zero.
        AuxMat = Delta * 0.5 * (np.dot(2.*B - A, DD0) + A)

        # a --> Define the random vector x0 for the factors. Note that x0 takes
        # value in the hypercube
        # [0,...,1-Delta]*[0,...,1-Delta]*[0,...,1-Delta]*[0,...,1-Delta]
        # Original in Stijn Van Hoey's version
        # xset=np.arange(0.0,1.0-Delta,1.0/(p-1))
        # Jule's version from The Primer
        # xset=np.arange(0.0,1.0-1.0/(p-1),1.0/(p-1))
        # Matthias thinks that the difference between Python and Matlab is that
        # Python is not taking the last element; therefore the following
        # version
        xset = np.arange(0.0, 1.00000001 - Delta, 1.0 / (p - 1))
        x0 = xset.take(list(np.ceil(
            np.random.random(k) * np.floor(p / 2)) - 1))
        x0 = x0[np.newaxis, :]

        # b --> Compute the matrix B*, here indicated as B0. Each row in B0 is
        # a trajectory for Morris Calculations. The dimension of B0 is
        # (Numfactors+1,Numfactors)
        if Groupnumber != 0:
            B0 = np.dot(A0, x0) + AuxMat
        else:
            B0 = np.dot(np.dot(A0, x0) + AuxMat, P0)

        # c --> Compute values in the original intervals
        # B0 has values x(i,j) in [0, 1/(p -1), 2/(p -1), ... , 1].
        # To obtain values in the original intervals [lb, ub] we compute
        # lb(j) + x(i,j)*(ub(j)-lb(j))
        In = np.tile(lb, (sizeb, 1)) + B0 * np.tile((ub - lb), (sizeb, 1))

        # Create the Factor vector. Each component of this vector indicate
        # which factor or group of factor has been changed in each step of the
        # trajectory.
        for j in range(sizea):
            Fact[j] = np.where(P0[j, :])[0]
        # Enkel om vorm logisch te houden. of Fact kleiner maken
        Fact[sizea] = int(-1)

        # append the create traject to the others
        Outmatrix[i*(sizea+1):(i+1)*(sizea+1), :] = In
        OutFact[i*(sizea+1):(i+1)*(sizea+1)] = Fact.reshape((sizea+1, 1))

    return Outmatrix, OutFact


def Optimized_Groups(NumFact, lb, ub, r,
                     p=6, N=None,
                     dist=None, distparam=None,
                     GroupMat=np.array([]), Diagnostic=0):
    """
    Optimisation in the choice of trajectories for Morris experiment,
    that means elementary effects


    Definition
    ----------
    def Optimized_Groups(NumFact, lb, ub, r, p=6, N=None,
                         GroupMat=np.array([]), Diagnostic=0):


    Input
    -----
    NumFact
        Number of factors
    lb
        [NumFact] Lower bound of the uniform distribution for each factor
        or lower fraction of percent point function ppf if distribution given.
    ub
        [NumFact] Upper bound of the uniform distribution for each factor
        or upper fraction of percent point function ppf if distribution given.
    r
        Final number of optimal trajectories


    Optional Input
    --------------
    p
        Number of levels (default: 6)
    N
        Total number of trajectories. If None: N=10*r (default: None)
    dist
        List of None or scipy.stats distribution objects for each factor
        having the method ppf, Percent Point Function (Inverse of CDF)
        (default: None).

        If None, the uniform distribution will be sampled from lower bound lb
        to upper bound ub.

        If dist is scipy.stats.uniform, the ppf will be sampled from the lower
        fraction given in lb and the upper fraction in ub. The sampling
        interval is then given by the parameters loc=lower and
        scale=interval=upper-lower in param.
    distparam
        List with tuples with parameters as required for dist (default: (0,1)).
        All distributions of scipy.stats have location and scale parameters,
        at least. loc and scale are implemented as keyword arguments in
        scipy.stats. Other parameters such as the shape parameter of the gamma
        distribution must hence be given first,
        e.g. (shape,loc,scale) for the gamma distribution.
        distparam is ignored if dist is None.
        The percent point function ppf is called like this:
        dist(*distparam).ppf(x)
    GroupMat
        [NumFact,NumGroups] Matrix describing the groups.
        (default: np.array([]))

        Each column represents a group and its elements are set to 1 in
        correspondence of the factors that belong to the fixed group. All the
        other elements are zero.
    Diagnostic
        1=plot the histograms and compute the efficiency of the samplign or
        not, 0 otherwise (default)


    Output
    ------
    [OptMatrix, OptOutVec]


    References
    ----------
    Saltelli, A., Chan, K., & Scott, E. M. (2000). Sensitivity Analysis. Wiley
    Series in Probability and Statistics, John Wiley & Sons, New York, 1-504. -
    on page 68ff

    """
    from scipy.spatial import distance
    import scipy.stats as stats

    if N is None:
        N = 10*r

    assert len(lb) == NumFact, 'Lower bound must have length NumFact.'
    assert len(ub) == NumFact, 'Upper bound must have length NumFact.'
    if dist is not None:
        assert len(lb) == len(dist), 'scipy.stats distribution object or None has to be given for each parameter.'
        for dd in dist:
            if dd is not None:
                if not isinstance(dd, (stats.rv_discrete, stats.rv_continuous)):
                    raise TypeError(str(dd)+' is not a scipy.stats distribution object.')

    # np.random.seed(seed=1025)
    # Sample trajectorie between 0 and 1. Will be rescaled to specific
    # distributions later.
    lbt = np.zeros(NumFact)
    ubt = np.ones(NumFact)
    # Version with Groups
    OutMatrix, OutFact = Sampling_Function_2(p, NumFact, N, lbt, ubt, GroupMat)

    try:
        Groupnumber = GroupMat.shape[1]
    except:
        Groupnumber = 0

    if Groupnumber != 0:
        sizeb = Groupnumber + 1
    else:
        sizeb = NumFact + 1

    # Compute the distance between all pair of trajectories (sum of the
    # distances between points) The distance matrix is a matrix N*N The
    # distance is defined as the sum of the distances between all pairs of
    # points if the two trajectories differ, 0 otherwise
    Dist = np.zeros((N, N))
    Diff_Traj = np.arange(0.0, N, 1.0)
    # combine all trajectories: eg N=3: 0&1; 0&2; 1&2 (is not dependent from
    # sequence)
    for j in range(N):
        for z in range(j+1, N):
            MyDist = distance.cdist(OutMatrix[sizeb*j:sizeb*(j+1), :],
                                    OutMatrix[sizeb*z:sizeb*(z+1), :])
            if np.where(MyDist == 0.)[0].size == sizeb:
                # Same trajectory. If the number of zeros in Dist matrix is
                # equal to (NumFact+1) then the trajectory is a replica. In
                # fact (NumFact+1) is the maximum number of points that two
                # trajectories can have in common
                Dist[j, z] = 0.
                Dist[z, j] = 0.
                # Memorise the replicated trajectory
                Diff_Traj[z] = -1.  # the z value identifies the duplicate
            else:
                # Define the distance between two trajectories as
                # the minimum distance among their points
                dd = np.sum(MyDist)
                Dist[j, z] = dd
                Dist[z, j] = dd

    # prepare array with excluded duplicates (alternative would be deleting
    # rows)
    iidup = np.where(Diff_Traj == -1.)[0]
    dupli = iidup.size
    iiind = np.where(Diff_Traj != -1.)[0]
    New_N = iiind.size  # N - iidup.size
    New_OutMatrix = np.zeros((sizeb*New_N, NumFact))
    New_OutFact   = np.zeros((sizeb*New_N, 1))

    # Eliminate replicated trajectories in the sampled matrix
    ID = 0
    for i in range(N):
        if Diff_Traj[i] != -1.:
            New_OutMatrix[ID*sizeb:(ID+1)*sizeb, :] = (
                OutMatrix[i*sizeb:(i+1)*sizeb, :])
            New_OutFact[ID*sizeb:(ID+1)*sizeb, :] = (
                OutFact[i*sizeb:(i+1)*sizeb, :])
            ID += 1

    # Select in the distance matrix only the rows and columns of different
    #   trajectories Dist_Diff =
    #   np.delete(Dist_Diff,np.where(Diff_Traj==-1.)[0])
    # moet 2D matrix zijn... wis rijen ipv hou bij
    Dist_Diff = Dist[iiind, :]
    Dist_Diff = Dist_Diff[:, iiind]

    # Select the optimal set of trajectories
    Traj_Vec = np.zeros((New_N, r), dtype=int)
    OptDist  = np.zeros((New_N, r))
    for m in range(New_N):  # each row in Traj_Vec
        Traj_Vec[m, 0] = m
        for z in range(1, r):  # elements in columns after first
            New_Dist_Diff  = np.sqrt(np.sum(Dist_Diff[Traj_Vec[m, :z], :]**2,
                                            axis=0))
            ii = New_Dist_Diff.argmax()
            Traj_Vec[m, z] = ii
            OptDist[m, z]  = New_Dist_Diff[ii]

    # Construct optimal matrix
    SumOptDist = np.sum(OptDist, axis=1)
    # Find the maximum distance
    Pluto = np.where(SumOptDist == SumOptDist.max())[0]
    Opt_Traj_Vec = Traj_Vec[Pluto[0], :]

    OptMatrix = np.zeros((sizeb*r, NumFact))
    OptOutVec = np.zeros((sizeb*r, 1))

    for k in range(r):
        OptMatrix[k*sizeb:(k+1)*sizeb, :] = (
            New_OutMatrix[sizeb*Opt_Traj_Vec[k]:sizeb*(Opt_Traj_Vec[k]+1), :])
        OptOutVec[k*sizeb:(k+1)*sizeb, :] = (
            New_OutFact[sizeb*Opt_Traj_Vec[k]:sizeb*(Opt_Traj_Vec[k]+1), :])

    # ----------------------------------------------------------------------
    # Compute values in the original intervals
    # Optmatrix has values x(i,j) in [0, 1/(p -1), 2/(p -1), ... , 1].
    # To obtain values in the original intervals [lb, ub] we compute
    # lb(j) + x(i,j)*(ub(j)-lb(j))
    if Diagnostic:
        OptMatrix_b = OptMatrix.copy()  # save for plot
    if dist is None:
        OptMatrix = (np.tile(lb, (sizeb*r, 1)) +
                     np.tile(ub-lb, (sizeb*r, 1)) * OptMatrix)
    else:
        for i, dd in enumerate(dist):
            OptMatrix[:, i] = lb[i] + (ub[i]-lb[i]) * OptMatrix[:, i]
            if dd is not None:
                if distparam is None:
                    pars = (0., 1.)
                else:
                    if distparam[i] is None:
                        pars = (0., 1.)
                    else:
                        pars = tuple([ float(k) for k in distparam[i] ])
                OptMatrix[:, i] = dd(*pars).ppf(OptMatrix[:, i])

    if Diagnostic:
        # Clean the trajectories from repetitions and plot the histograms
        hplot = np.zeros((2 * r, NumFact))

        for i in range(NumFact):
            for j in range(r):
                # select the first value of the factor
                hplot[j * 2, i] = OptMatrix_b[j * sizeb, i]

                # search the second value
                for ii in range(1, sizeb):
                    if OptMatrix_b[j*sizeb+ii, i] != OptMatrix_b[j*sizeb, i]:
                        kk = 1
                        hplot[j*2+kk, i] = OptMatrix_b[j*sizeb+ii, i]

        try:  # pragma: no cover
            import matplotlib as mpl
            mpl.use('Agg')
            import matplotlib.pyplot as plt
            mpl.rcParams['font.family'] = 'sans-serif'
            mpl.rcParams['font.sans-serif'] = 'Arial'  # Arial, Verdana
            mpl.rc('savefig', dpi=300, format='png')
            mpl.rc('font', size=9)
            fig = plt.figure()
            fig.suptitle('New Strategy')
            DimPlots = NumFact // 2 + (1 if NumFact % 2 > 0 else 0)
            for i in range(NumFact):
                ax = fig.add_subplot(DimPlots, 2, i + 1)
                ax.hist(hplot[:, i], p)
            fig.savefig('morris_diag_new_strategy.png', transparent=False,
                        bbox_inches='tight', pad_inches=0.035)
            plt.close(fig)
        except ImportError:  # pragma: no cover
            pass

        # Plot the histogram for the original sampling strategy
        # Select the matrix
        OrigSample = OutMatrix[:r*(sizeb), :]
        Orihplot   = np.zeros((2*r, NumFact))

        for i in range(NumFact):
            for j in range(r):
                # select the first value of the factor
                Orihplot[j * 2, i] = OrigSample[j * sizeb, i]

                # search the second value
                for ii in range(1, sizeb):
                    if OrigSample[j * sizeb + ii, i] != OrigSample[j *
                                                                   sizeb, i]:
                        kk = 1
                        Orihplot[j * 2 + kk, i] = OrigSample[j * sizeb + ii, i]

        try:  # pragma: no cover
            fig = plt.figure()
            fig.suptitle('Old Strategy')
            DimPlots = NumFact // 2 + (1 if NumFact % 2 > 0 else 0)
            for i in range(NumFact):
                ax = fig.add_subplot(DimPlots, 2, i + 1)
                ax.hist(Orihplot[:, i], p)
                # plt.title('Old Strategy')
            fig.savefig('morris_diag_old_strategy.png', transparent=False,
                        bbox_inches='tight', pad_inches=0.035)
            plt.close(fig)
        except:  # pragma: no cover
            pass

        # Measure the quality of the sampling strategy
        levels = np.arange(0.0, 1.1, 1.0 / (p - 1))
        NumSPoint = np.zeros((NumFact, p))
        NumSOrigPoint = np.zeros((NumFact, p))
        for i in range(NumFact):
            for j in range(p):
                # For each factor and each level count the number of times the
                # factor is on the level
                # This for the new and original sampling
                NumSPoint[i, j] = np.where(
                    np.abs(hplot[:, i] - np.tile(levels[j], hplot.shape[0]))
                    < 1e-5)[0].size
                NumSOrigPoint[i, j] = np.where(
                    np.abs(Orihplot[:, i] -
                           np.tile(levels[j], Orihplot.shape[0]))
                    < 1e-5)[0].size

        # The optimal sampling has values uniformly distributed across the
        # levels
        OptSampl       = 2. * r / p
        QualMeasure    = 0.
        QualOriMeasure = 0.
        for i in range(NumFact):
            for j in range(p):
                QualMeasure = QualMeasure + np.abs(NumSPoint[i, j] - OptSampl)
                QualOriMeasure = QualOriMeasure + np.abs(
                    NumSOrigPoint[i, j] - OptSampl)

        QualMeasure = 1. - QualMeasure / (OptSampl * p * NumFact)
        QualOriMeasure = 1. - QualOriMeasure / (OptSampl * p * NumFact)

        print('The quality of the sampling strategy changed from {:f} with '
              'the old strategy to {:f} '
              'for the optimized strategy'.format(QualOriMeasure, QualMeasure))

    return OptMatrix, OptOutVec[:, 0]


def Morris_Measure_Groups(NumFact, Sample, OutFact, Output, p=4,
                          Group=[], Diagnostic=False):
    """
    Given the Morris sample matrix, the output values and the group matrix
    compute the Morris measures.


    Definition
    ----------
    def Morris_Measure_Groups(NumFact, Sample, OutFact, Output, p=4, Group=[],
                              Diagnostic=False):


    Input
    -----
    NumFact
        Number of factors
    Sample
        Matrix of the Morris sampled trajectories
    OutFact
        Matrix with the factor changings as specified in Morris sampling
    Output
        Matrix of the output(s) values in correspondence of each point of each
        trajectory


    Optional Input
    --------------
    p
        Number of levels
    Group
        [NumFact, NumGroups] Matrix describing the groups.
        Each column represents one group. The element of each column are zero
        if the factor is not in the group. Otherwise it is 1.
    Diagnostic
        True:  print out diagnostics

        False: otherwise (default)


    Output
    ------
    SA, OutMatrix
        SA(NumFact*Output.shape[1],N) individual sensitivity measures

        OutMatrix(NumFact*Output.shape[1], 3) = [Mu*, Mu, StDev]
        Morris Measures
        It gives the three measures of each factor for each output.


    References
    ----------
    Saltelli, A., Chan, K., & Scott, E. M. (2000). Sensitivity Analysis. Wiley
    Series in Probability and Statistics, John Wiley & Sons, New York, 1-504. -
    on page 68ff

    """
    try:
        NumGroups = Group.shape[1]
        if Diagnostic:
            print('{:d} Groups are used'.format(NumGroups))
    except:
        NumGroups = 0
        if Diagnostic:
            print('No Groups are used')

    Delt = p / (2. * (p - 1.))

    if NumGroups != 0:
        sizea = NumGroups
        GroupMat = Group
        GroupMat = GroupMat.transpose()
        if Diagnostic:
            print('NumGroups', NumGroups)
    else:
        sizea = NumFact
    sizeb = sizea + 1

    # r = Sample.shape[0]/sizeb
    r = Sample.shape[0] // sizeb

    try:
        NumOutp = Output.shape[1]
    except:
        NumOutp = 1
        Output = Output.reshape((Output.size, 1))

    # For each Output
    if NumGroups == 0:
        # for every output: every factor is a line, columns are mu*, mu and std
        OutMatrix = np.zeros((NumOutp * NumFact, 3))
    else:
        # for every output: every factor is a line, column is mu*
        OutMatrix = np.zeros((NumOutp * NumFact, 1))

    SAmeas_out = np.zeros((NumOutp * NumFact, r))

    for k in range(NumOutp):
        OutValues = Output[:, k]

        # For each trajectory
        # vorm afhankelijk maken van group of niet...
        SAmeas = np.zeros((NumFact, r))
        for i in range(r):
            # For each step j in the trajectory
            # Read the orientation matrix fact for the r-th sampling
            # Read the corresponding output values
            # Read the line of changing factors
            Single_Sample = Sample[i * sizeb:(i + 1) * sizeb, :]
            Single_OutValues = OutValues[i * sizeb:(i + 1) * sizeb]
            # gives factor in change (or group)
            Single_Facts = np.array(OutFact[i * sizeb:(i + 1) * sizeb],
                                    dtype=int)

            A = (Single_Sample[1:sizeb, :] -
                 Single_Sample[:sizea, :]).transpose()
            Delta = A[np.where(A)]  # AAN TE PASSEN?
            # If upper bound==lower bound then A==0 in all trajectories. Delta
            # will have not the right dimensions then because these are
            # filtered out with where. Fill in Delta==0 for these cases.
            ii = np.where(np.sum(A, axis=0) == 0.)[0]
            if ii.size > 0:
                Delta = np.insert(Delta, ii, 0.)

            if Diagnostic:
                print('A: ', A)
                print('Delta: ', Delta)
                print('Single_Facts: ', Single_Facts)

            # For each point of the fixed trajectory, i.e. for each factor,
            # compute the values of the Morris function.
            for j in range(sizea):
                if NumGroups != 0:  # work with groups
                    Auxfind = A[:, j]
                    Change_factor = np.where(np.abs(Auxfind) > 1e-010)[0]
                    for gk in Change_factor:
                        SAmeas[gk, i] = np.abs(
                            (Single_OutValues[j] - Single_OutValues[j + 1]) /
                            Delt)  # nog niet volledig goe
                else:
                    if Delta[j] > 0.0:
                        SAmeas[Single_Facts[j], i] = ((Single_OutValues[j+1] -
                                                       Single_OutValues[j]) /
                                                      Delt)
                    else:
                        SAmeas[Single_Facts[j], i] = ((Single_OutValues[j] -
                                                       Single_OutValues[j + 1])
                                                      / Delt)

        # Compute Mu AbsMu and StDev
        if np.isnan(SAmeas).any():
            AbsMu = np.zeros(NumFact)
            Mu = np.zeros(NumFact)
            Stdev = np.zeros(NumFact)

            for j in range(NumFact):
                SAm = SAmeas[j, :]
                SAm = SAm[~np.isnan(SAm)]
                rr = float(SAm.size)
                AbsMu[j] = np.sum(np.abs(SAm)) / rr
                if NumGroups == 0:
                    Mu[j] = SAm.mean()
                    if SAm.size > 1:
                        Stdev[j] = np.std(SAm, ddof=1)
                    else:
                        Stdev[j] = 0.
        else:
            AbsMu = np.sum(np.abs(SAmeas), axis=1) / r
            if NumGroups == 0:
                Mu = SAmeas.mean(axis=1)
                if SAmeas.shape[1] > 1:
                    Stdev = np.std(SAmeas, ddof=1, axis=1)
                else:
                    Stdev = np.zeros(SAmeas.shape[0])
            else:
                Mu = np.zeros(NumFact)
                Stdev = np.zeros(NumFact)

        OutMatrix[k * NumFact:k * NumFact + NumFact, 0] = AbsMu
        if NumGroups == 0:
            OutMatrix[k * NumFact:k * NumFact + NumFact, 1] = Mu
            OutMatrix[k * NumFact:k * NumFact + NumFact, 2] = Stdev

        SAmeas_out[k * NumFact:k * NumFact + NumFact, :] = SAmeas

    return SAmeas_out, OutMatrix


def morris_sampling(nparam, lb, ub, nt,
                    nsteps=6, ntotal=None,
                    dist=None, distparam=None,
                    GroupMat=np.array([]), Diagnostic=0):
    """
    Sample trajectories in parameter space

    Optimisation in the choice of trajectories for Morris experiment, that
    means elementary effects.

    Parameters
    ----------
    nparam : int
        Number of parameters / factors
    lb : array_like
        (nparam,) Lower bound of the uniform distribution for each parameter /
        factor or lower fraction of percent point function ppf if distribution
        given.
        Be aware that the percent point function *ppf* of most continuous
        distributions is infinite at 0.
    ub : array_like
        (nparam,) Upper bound of the uniform distribution for each parameter /
        factor or upper fraction of percent point function ppf if distribution
        given.
        Be aware that the percent point function *ppf* of most continuous
        distributions is infinite at 1.
    nt : int
        Final number of optimal trajectories
    nsteps : int, optional
        Number of levels, i.e. intervals in trajectories (default: 6)
    ntotal : int, optional
        Total number of sampled trajectories. If None: `ntotal=10*nt` (default:
        None)
    dist : list, optional
        List of None or scipy.stats distribution objects for each factor having
        the method ppf, Percent Point Function (Inverse of CDF) (default: None)
        If None, the uniform distribution will be sampled from lower bound `lb`
        to upper bound `ub`.
        If `dist` is scipy.stats.uniform, the ppf will be sampled from the
        lower fraction given in `lb` and the upper fraction in `ub`. The
        sampling interval is then given by the parameters `loc=lower` and
        `scale=interval=upper-lower` in `distparam`. This means
        `dist=None`, `lb=a`, `ub=b`
        corresponds to
        `lb=0`, `ub=1`, `dist=scipy.stats.uniform`, `distparam=[a,b-a]`
    distparam : list, optional
        List with tuples with parameters as required for `dist` (default:
        (0,1)).
        All distributions of scipy.stats have location and scale parameters, at
        least. `loc` and `scale` are implemented as keyword arguments in
        scipy.stats. Other parameters such as the shape parameter of the gamma
        distribution must hence be given first, e.g. `(shape,loc,scale)` for
        the gamma distribution.
        `distparam` is ignored if `dist` is None.
        The percent point function ppf is called like this:
        `dist(*distparam).ppf(x)`
    GroupMat : ndarray, optional
        (nparam,ngroup) Matrix describing the groups. (default: np.array([]))
        Each column represents a group. The elements of each column are zero
        if the parameter / factor is not in the group, otherwise it is 1.
    Diagnostic : bool, optional
        Plot the histograms and compute the efficiency of the sampling if True
        (default: False)

    Returns
    -------
    traj : list
        list [OptMatrix, OptOutVec] with `OptMatrix((nparam+1)*nt,nparam)` and
        `OptOutVec(nparam*nt)`

    References
    ----------
    Saltelli, A., Chan, K., & Scott, E. M. (2000). Sensitivity Analysis. Wiley
    Series in Probability and Statistics, John Wiley & Sons, New York, 1-504. -
    on page 68ff

    Examples
    --------
    >>> import numpy as np
    >>> seed = 1023
    >>> np.random.seed(seed=seed)
    >>> npara = 10
    >>> x0    = np.ones(npara)*0.5
    >>> lb    = np.zeros(npara)
    >>> ub    = np.ones(npara)
    >>> mask  = np.ones(npara, dtype=bool)
    >>> mask[5::2] = False
    >>> nmask = np.sum(mask)
    >>> nt     = npara
    >>> ntotal = max(nt**2, 10*nt)
    >>> nsteps = 6
    >>> tmatrix, tvec = morris_sampling(nmask, lb[mask], ub[mask], nt,
    ...                                 nsteps=nsteps, ntotal=ntotal,
    ...                                 Diagnostic=False)
    >>> # Set input vector to trajectories and masked elements to x0
    >>> x = np.tile(x0, tvec.size).reshape(tvec.size, npara)  # default to x0
    >>> x[:,mask] = tmatrix  # replaced unmasked with trajectorie values
    >>> print(x[0,:])
    [0.6 0.4 0.8 0.6 0.6 0.5 0.4 0.5 0.  0.5]

    >>> import scipy.stats as stats
    >>> seed = 1023
    >>> np.random.seed(seed=seed)
    >>> npara = 10
    >>> x0    = np.ones(npara)*0.5
    >>> lb    = np.zeros(npara)
    >>> ub    = np.ones(npara)
    >>> dist  = [ stats.uniform for i in range(npara) ]
    >>> distparam = [ (lb[i], ub[i]-lb[i]) for i in range(npara) ]
    >>> mask  = np.ones(npara, dtype=bool)
    >>> mask[5::2] = False
    >>> nmask = np.sum(mask)
    >>> nt     = npara
    >>> ntotal = max(nt**2, 10*nt)
    >>> nsteps = 6
    >>> tmatrix, tvec = morris_sampling(nmask, lb[mask], ub[mask], nt,
    ...                                 nsteps=nsteps, ntotal=ntotal,
    ...                                 dist=dist, distparam=distparam,
    ...                                 Diagnostic=False)
    >>> # Set input vector to trajectories and masked elements to x0
    >>> x = np.tile(x0, tvec.size).reshape(tvec.size, npara)  # default to x0
    >>> x[:,mask] = tmatrix  # replaced unmasked with trajectory values
    >>> print(x[0,:])
    [0.6 0.4 0.8 0.6 0.6 0.5 0.4 0.5 0.  0.5]

    """
    return Optimized_Groups(nparam, lb, ub, nt,
                            p=nsteps, N=ntotal,
                            dist=dist, distparam=distparam,
                            GroupMat=GroupMat, Diagnostic=Diagnostic)


def elementary_effects(nparam, OptMatrix, OptOutVec, Output,
                       nsteps=4, Group=[], Diagnostic=False):
    """
    Compute the Morris measures given the Morris sample matrix, the output
    values and the group matrix.

    Parameters
    ----------
    nparam : int
        Number of parameters / factors list [OptMatrix, OptOutVec] with
        OptMatrix((nparam+1)*nt,nparam) and OptOutVec(nparam*nt)
    OptMatrix : ndarray
        ((nparam+1)*nt,nparam) Matrix of the Morris sampled trajectories from
        morris_sampling
    OptOutVec : ndarray
        (nparam*nt,) Matrix with the parameter / factor changings from
        morris_sampling
    Output : ndarray
        ((nparam+1)*nt,) Matrix of the output values of each point of each
        trajectory
    nsteps : int, optional
        Number of levels, i.e. intervals in trajectories (default: 4)
    Group : ndarray, optional
        (nparam,NumGroups) Matrix describing the groups. (default: [])
        Each column represents a group. The elements of each column are zero
        if the parameter / factor is not in the group, otherwise it is 1.
    Diagnostic : bool, optional
        Print out diagnostics if True (default: False)

    Returns
    -------
    SA, OutMatrix : list of ndarrays
        SA(nparam*Output.shape[1],N) individual sensitivity measures,
        OutMatrix(nparam*Output.shape[1], 3) = [Mu*, Mu, StDev] Morris Measures

        It gives the three measures of each parameter / factor for each output.

    References
    ----------
    Saltelli, A., Chan, K., & Scott, E. M. (2000). Sensitivity Analysis. Wiley
    Series in Probability and Statistics, John Wiley & Sons, New York, 1-504. -
    on page 68ff

    Examples
    --------
    >>> import numpy as np
    >>> seed = 1023
    >>> np.random.seed(seed=seed)
    >>> npara = 10
    >>> x0    = np.ones(npara)*0.5
    >>> lb    = np.zeros(npara)
    >>> ub    = np.ones(npara)
    >>> mask  = np.ones(npara, dtype=bool)
    >>> mask[5::2] = False
    >>> nmask = np.sum(mask)
    >>> nt     = npara
    >>> ntotal = max(nt**2, 10*nt)
    >>> nsteps = 6
    >>> tmatrix, tvec = morris_sampling(nmask, lb[mask], ub[mask], nt,
    ...                                 ntotal=ntotal, nsteps=nsteps,
    ...                                 Diagnostic=False)
    >>> # Set input vector to trajectories and masked elements to x0
    >>> x = np.tile(x0, tvec.size).reshape(tvec.size, npara)  # default to x0
    >>> x[:,mask] = tmatrix  # replaced unmasked with trajectorie values
    >>> func = np.sum
    >>> fx = np.array(list(map(func,x)))
    >>> out = np.zeros((npara,3))
    >>> sa, res = elementary_effects(nmask, tmatrix, tvec, fx, nsteps=nsteps,
    ...                              Diagnostic=False)
    >>> out[mask,:] = res
    >>> print(out[:,0])
    [1. 1. 1. 1. 1. 0. 1. 0. 1. 0.]

    """
    return Morris_Measure_Groups(nparam, OptMatrix, OptOutVec, Output, nsteps, Group, Diagnostic)


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # # nt = 10
    # print('nt=10')
    # np.random.seed(seed=1234)
    # nparam = 15
    # lb = np.arange(nparam)
    # ub = 2.*lb + 1.
    # nt     = 10
    # nsteps = 6
    # ntotal = 100
    # Diagnostic = 0
    # out = np.random.random(nt*(nparam+1))
    # mat, vec = morris_sampling(nparam, lb, ub, nt, nsteps=nsteps, ntotal=ntotal, Diagnostic=Diagnostic)
    # print(np.around(mat[0,0:5],3))
    # print(np.around(vec[0:5],3))
    # sa, res = elementary_effects(nparam, mat, vec, out, nsteps=nsteps)
    # print(np.around(res[0:5,0],3)) # (nparam,3) = AbsMu, Mu, Stddev
    # print(np.around(sa[0:5,1],3))  #  (nparam,r) individual elementary effects for all parameters

    # # nt = 10, nan
    # print('nt=10, nan')
    # np.random.seed(seed=1234)
    # nparam = 15
    # lb = np.arange(nparam)
    # ub = 2.*lb + 1.
    # nt     = 10
    # nsteps = 6
    # ntotal = 100
    # Diagnostic = 0
    # out = np.random.random(nt*(nparam+1))
    # out[1:nt*nparam:nparam//2] = np.nan
    # mat, vec = morris_sampling(nparam, lb, ub, nt, nsteps=nsteps, ntotal=ntotal, Diagnostic=Diagnostic)
    # print(np.around(mat[0,0:5],3))
    # print(np.around(vec[0:5],3))
    # sa, res = elementary_effects(nparam, mat, vec, out, nsteps=nsteps)
    # print(np.around(res[0:5,0],3)) # (nparam,3) = AbsMu, Mu, Stddev
    # print(np.around(sa[~np.isnan(sa[:,1]),1],3))  #  (nparam,r) individual elementary effects for all parameters

    # # nt = 1
    # print('nt=1')
    # np.random.seed(seed=1234)
    # nparam = 15
    # lb = np.arange(nparam)
    # ub = 2.*lb + 1.
    # nt     = 1
    # nsteps = 6
    # ntotal = 100
    # Diagnostic = 0
    # out = np.random.random(nt*(nparam+1))
    # mat, vec = morris_sampling(nparam, lb, ub, nt, nsteps=nsteps, ntotal=ntotal, Diagnostic=Diagnostic)
    # sa, res = elementary_effects(nparam, mat, vec, out, nsteps=nsteps)
    # print(np.around(res[0:5,0],3)) # (nparam,3) = AbsMu, Mu, Stddev
    # print(np.around(sa[0:5].squeeze(),3))  #  (nparam,r) individual elementary effects for all parameters

    # # groups
    # print('groups')
    # np.random.seed(seed=1234)
    # nparam   = 15
    # ngroup = 5
    # lb = np.arange(nparam)
    # ub = 2.*lb + 1.
    # Groups = np.random.randint(0, 4, (nparam,ngroup))
    # nt     = 10
    # nsteps = 6
    # ntotal = 100
    # Diagnostic = 0
    # out = np.random.random(nt*(nparam+1))
    # mat, vec = morris_sampling(nparam, lb, ub, nt, nsteps=nsteps, ntotal=ntotal, GroupMat=Groups, Diagnostic=Diagnostic)
    # print(np.around(mat[0,0:5],3))
    # print(np.around(vec[0:5],3))
    # sa, res = elementary_effects(nparam, mat, vec, out, nsteps=nsteps, Group=Groups)
    # print(np.around(res[0:5,0],3)) # (nparam,3) = AbsMu, Mu, Stddev
    # print(np.around(sa[0:5,1],3))  #  (nparam,r) individual elementary effects for all parameters

    # # groups, nan
    # print('groups, nan')
    # np.random.seed(seed=1234)
    # nparam   = 15
    # ngroup = 5
    # lb = np.arange(nparam)
    # ub = 2.*lb + 1.
    # Groups = np.random.randint(0, 4, (nparam,ngroup))
    # nt     = 10
    # nsteps = 6
    # ntotal = 100
    # Diagnostic = 0
    # out = np.random.random(nt*(nparam+1))
    # out[1:nt*nparam:nparam//2] = np.nan
    # mat, vec = morris_sampling(nparam, lb, ub, nt, nsteps=nsteps, ntotal=ntotal, GroupMat=Groups, Diagnostic=Diagnostic)
    # print(np.around(mat[0,0:5],3))
    # print(np.around(vec[0:5],3))
    # sa, res = elementary_effects(nparam, mat, vec, out, nsteps=nsteps, Group=Groups)
    # print(np.around(res[0:5,0],3)) # (nparam,3) = AbsMu, Mu, Stddev
    # print(np.around(sa[0:5,1],3))  #  (nparam,r) individual elementary effects for all parameters

    # # nt=10, dist
    # print('nt=10, distributions')
    # import scipy.stats as stats
    # np.random.seed(seed=1234)
    # nparam = 15
    # lb = np.arange(nparam)
    # ub = 2.*lb + 1.
    # lb = np.zeros(nparam)
    # ub = np.ones(nparam)
    # dist      = [ stats.uniform for i in range(nparam) ]
    # distparam = [ (lb[i],ub[i]-lb[i]) for i in range(nparam) ]
    # lb = np.zeros(nparam)
    # ub = np.ones(nparam)
    # nt     = 10
    # nsteps = 6
    # ntotal = 100
    # Diagnostic = 0
    # out = np.random.random(nt*(nparam+1))
    # mat, vec = morris_sampling(nparam, lb, ub, nt, nsteps=nsteps, ntotal=ntotal, dist=dist, distparam=distparam, Diagnostic=Diagnostic)
    # print(np.around(mat[0,0:5],3))
    # print(np.around(vec[0:5],3))
    # sa, res = elementary_effects(nparam, mat, vec, out, nsteps=nsteps)
    # print(np.around(res[0:5,0],3)) # (nparam,3) = AbsMu, Mu, Stddev
    # print(np.around(sa[0:5,1],3))  #  (nparam,r) individual elementary effects for all parameters

    # # groups, dist
    # print('groups, distributions')
    # np.random.seed(seed=1234)
    # nparam   = 15
    # ngroup = 5
    # lb = np.arange(nparam)
    # ub = 2.*lb + 1.
    # dist      = [ stats.uniform for i in range(nparam) ]
    # distparam = [ (lb[i],ub[i]-lb[i]) for i in range(nparam) ]
    # lb = np.zeros(nparam)
    # ub = np.ones(nparam)
    # Groups = np.random.randint(0, 4, (nparam,ngroup))
    # nt     = 10
    # nsteps = 6
    # ntotal = 100
    # Diagnostic = 0
    # out = np.random.random(nt*(nparam+1))
    # mat, vec = morris_sampling(nparam, lb, ub, nt, nsteps=nsteps, ntotal=ntotal, dist=dist, distparam=distparam, GroupMat=Groups, Diagnostic=Diagnostic)
    # print(np.around(mat[0,0:5],3))
    # print(np.around(vec[0:5],3))
    # sa, res = elementary_effects(nparam, mat, vec, out, nsteps=nsteps, Group=Groups)
    # print(np.around(res[0:5,0],3)) # (nparam,3) = AbsMu, Mu, Stddev
    # print(np.around(sa[0:5,1],3))  #  (nparam,r) individual elementary effects for all parameters



#!/usr/bin/env python
"""
This is the unittest for tee module.

python -m unittest -v tests/test_tee.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_tee.py

"""
import unittest


class TestTee(unittest.TestCase):
    """
    Tests for tee.py
    """

    def test_tee(self):
        import os
        from pyjams import tee

        tee('T T T Test 1')
        ff = open('log.txt', 'w')
        tee('T T T Test 2', file=ff)
        ff.close()

        self.assertTrue(os.path.exists('log.txt'))

        ff = open('log.txt', 'r')
        inlog = ff.readline()
        ff.close()

        self.assertEqual(inlog.rstrip(), 'T T T Test 2')

        if os.path.exists('log.txt'):
            os.remove('log.txt')


if __name__ == "__main__":
    unittest.main()
